/***********************************************************************
* Driver for SP0A28 (0.3MP Camera)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
***********************************************************************/
//Gozone start BID xxxx add whole file
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/version.h>
#include <linux/vmalloc.h>
#include <linux/completion.h>
#include <media/v4l2-device.h>
#include <media/v4l2-subdev.h>
#include <linux/videodev2.h>
#include <media/v4l2-chip-ident.h>
#include <media/soc_camera.h>
#include <linux/videodev2_brcm.h>
#include <camdrv.h>
//#include <camdrv_SP0A28.h>
#include "camdrv_SP0A28.h"

#define SP0A28_NAME	"SP0A28"
#define SENSOR_ID 2
#define SENSOR_JPEG_SNAPSHOT_MEMSIZE	0x4b0000  /* 4915200 = 2560*1920 */ /*0x33F000*/ /*3403776*/ /*2216 * 1536*/
#define SENSOR_PREVIEW_WIDTH		640 /*  1024 */
#define SENSOR_PREVIEW_HEIGHT		480 /*  768 */

#define SP0A28_DEFAULT_PIX_FMT	V4L2_PIX_FMT_UYVY	/* YUV422 */
#define SP0A28_DEFAULT_MBUS_PIX_FMT	V4L2_MBUS_FMT_UYVY8_2X8
#define FORMAT_FLAGS_COMPRESSED		0x3
#define DEFUALT_MCLK			26000000

//#define SP0A28_REGISTER_SIZE		4  
#define SP0A28_REGISTER_SIZE		2    //modify by qiuzhoujun 20120925  

//zhuohengfeng++ for turnkey camera compatible
#define  SP0A28_MIPI_LANES 1
#define  SP0A28_MIPI_CHANNEL 1
#define  SP0A28_MIPI_HS_TERM_TIME 1



extern  int camdrv_i2c_set_config_register(struct i2c_client *client,
		const struct sensor_reg reg_buffer[],
		int num_of_regs,
		char *name);


/* #define __JPEG_CAPTURE__ 1 */       /* denis_temp ; yuv capture */

static const struct camdrv_framesize SP0A28_supported_preview_framesize_list[] = {
	{PREVIEW_SIZE_VGA,	640,  480 },
};


static const struct camdrv_framesize  SP0A28_supported_capture_framesize_list[] = {
	{CAPTURE_SIZE_VGA,	640,  480 },
};


const static struct v4l2_fmtdesc SP0A28_fmts[] = {
	{
		.index		= 0,
		.type		= V4L2_BUF_TYPE_VIDEO_CAPTURE,
		.flags		= 0,
		.description	= "UYVY",
		.pixelformat	= V4L2_MBUS_FMT_UYVY8_2X8,
	},

};

static const struct v4l2_queryctrl SP0A28_controls[] = {
	{
		.id		= V4L2_CID_CAMERA_BRIGHTNESS,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Brightness",
		.minimum	= EV_MINUS_2,
		.maximum	= EV_PLUS_2,
		.step		= 1,
		.default_value	= EV_DEFAULT,
	},

	{
		.id		= V4L2_CID_CAMERA_WHITE_BALANCE,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Whilte Balance",
		.minimum	= WHITE_BALANCE_AUTO ,
		.maximum	= WHITE_BALANCE_FLUORESCENT,
		.step		= 1,
		.default_value	= WHITE_BALANCE_AUTO,
	},


	{
		.id		= V4L2_CID_CAMERA_EFFECT,
		.type		= V4L2_CTRL_TYPE_INTEGER,
		.name		= "Color Effects",
		.minimum	= IMAGE_EFFECT_NONE,
		.maximum	= (1 << IMAGE_EFFECT_NONE |
					1 << IMAGE_EFFECT_BNW |
					1 << IMAGE_EFFECT_SEPIA |
					1 << IMAGE_EFFECT_NEGATIVE), /* this should be replace by querymenu */
		.step		= 1,
		.default_value	= IMAGE_EFFECT_NONE,
	},

	{
		.id		= V4L2_CID_CAMERA_CAPTURE,
		.type		= V4L2_CTRL_TYPE_BOOLEAN,
		.name		= "Capture",
		.minimum	= 0,
		.maximum	= 0,
		.step		= 0,
		.default_value	= 0,
	},
	
};


//===================define the SP0A28 function==============================================



//==========================================================================================
static int camdrv_SP0A28_enum_frameintervals(
	struct v4l2_subdev *sd,
	struct v4l2_frmivalenum *fival)
{
	int err = 0;
	int size, i;

	if (fival->index >= 1)
		return -EINVAL;

	CAM_INFO_MSG("check w = %d h = %d\n",fival->width, fival->height);
    
	fival->type = V4L2_FRMIVAL_TYPE_DISCRETE;

	for (i = 0; i < ARRAY_SIZE(SP0A28_supported_preview_framesize_list); i++) {
		if ((SP0A28_supported_preview_framesize_list[i].width == fival->width) &&
			(SP0A28_supported_preview_framesize_list[i].height == fival->height)) {
			size = SP0A28_supported_preview_framesize_list[i].index;
			break;
		}
	}

	if (i == ARRAY_SIZE(SP0A28_supported_preview_framesize_list)) {
		CAM_INFO_MSG(" unsupported width = %d and height = %d\n",
			 fival->width, fival->height);
		return -EINVAL;
	} else {
		CAM_INFO_MSG(" found i = %d\n", i);
	}

	switch (size) {
	case PREVIEW_SIZE_5MP:
	case PREVIEW_SIZE_3MP:
	case PREVIEW_SIZE_2MP:
	case PREVIEW_SIZE_W1MP:
	case PREVIEW_SIZE_1MP:
    case PREVIEW_SIZE_XGA: 
		fival->discrete.numerator = 2;
		fival->discrete.denominator = 15;
		break;
	default:
		fival->discrete.numerator = 1;
		fival->discrete.denominator = 30;
		break;
	}

	return err;
}


static long camdrv_SP0A28_ss_ioctl(
	struct v4l2_subdev *sd,
	unsigned int cmd, void *arg)
{
    int ret = 0;

    switch (cmd) {
    case VIDIOC_THUMB_SUPPORTED:
        {
            int *p = arg;
            *p = 0; /* no we don't support thumbnail */
            break;
        }
    case VIDIOC_JPEG_G_PACKET_INFO:
        {
            struct v4l2_jpeg_packet_info *p =
                (struct v4l2_jpeg_packet_info *)arg;
            p->padded = 0;
            p->packet_size = 0x400;
            break;
        }

    case VIDIOC_SENSOR_G_OPTICAL_INFO:
        {
            struct v4l2_sensor_optical_info *p =
                (struct v4l2_sensor_optical_info *)arg;
            /* assuming 67.5 degree diagonal viewing angle */
            p->hor_angle.numerator = 5401;
            p->hor_angle.denominator = 100;
            p->ver_angle.numerator = 3608;
            p->ver_angle.denominator = 100;
            p->focus_distance[0] = 10; /* near focus in cm */
            p->focus_distance[1] = 100; /* optimal focus in cm */
            p->focus_distance[2] = -1; /* infinity */
            p->focal_length.numerator = 342;
            p->focal_length.denominator = 100;
            break;
        }
    default:
        ret = -ENOIOCTLCMD;
        break;
    }
    return ret;
}




static int camdrv_SP0A28_i2c_set_data_burst(
	struct i2c_client *client,
	const struct sensor_reg reg_buffer[],
	int num_of_regs)
{
	//u16 subaddr = 0;  //modify by qiuzhoujun
	u8 subaddr = 0;  //add by qiuzhoujun 20120924
   // u16 data_value_16 = 0;
    u8  data_value_8 = 0;
	int i;
	int err = 0;

 	printk(KERN_ERR "in camdrv_SP0A28_i2c_set_data_burst and num_of_regs: %d  =======\n", num_of_regs);
	for (i = 0; i < num_of_regs; i++) {
        
        if(DELAY_SEQ == reg_buffer[i].dw){
            msleep(reg_buffer[i].val);
        }
        else{

    		//subaddr = reg_buffer[i].reg & 0xffff;  //disabled by qiuzhoujun 20120924
    		subaddr = reg_buffer[i].reg & 0xff;  //add by qiuzhoujun 20120924
    		data_value_8 = reg_buffer[i].val & 0xff;
		printk(KERN_ERR "in camdrv_SP0A28_i2c_set_data_burst and subaddr: 0x%x,   data: 0x%x     -----------------\n", subaddr, data_value_8);
		err = camdrv_i2c_write_2_bytes(client, subaddr, data_value_8);
#if 0
            if(8 == reg_buffer[i].dw){
                data_value_8 = reg_buffer[i].val & 0xff;
                //CAM_ERROR_MSG("[zhf]i2c write 3 byte: subaddr=0x%x, data_value_8=0x%x\n",subaddr, data_value_8);
                //err = camdrv_i2c_write_3_bytes(client, subaddr, data_value_8);
                err = camdrv_i2c_write_2_bytes(client, subaddr, data_value_8);  //modify by qiuzhoujun 20120924
            }
            else{
                data_value_16 = reg_buffer[i].val & 0xffff;
                //CAM_ERROR_MSG("[zhf]i2c write 4 byte: subaddr=0x%x, data_value_16=0x%x\n",subaddr,data_value_16);
                err = camdrv_i2c_write_4_bytes(client, subaddr, data_value_16);
            }  
#endif
		
            //if((err != 0)&&(subaddr != 0x001A)){  //modify by qiuzhoujun 20120924
            if((err != 0)&&(subaddr != 0xFD)){
                CAM_ERROR_MSG("i2c write fialed : subaddr=0x%x  data_value=0x%x ! \n",subaddr,data_value_8);
                break;
            }
        }
	}

    if (err < 0) {
        CAM_ERROR_MSG("i2c transfer failed ! \n");
        return -EIO;
    }

	return err;
}

static int camdrv_SP0A28_preview_setting(struct v4l2_subdev *sd)
{

    struct i2c_client *client = v4l2_get_subdevdata(sd);
    int err = 0;
#if 0
    // set SP0A28 to preview mode
    err = camdrv_i2c_set_config_register(client, SP0A28_preview_camera_regs, ARRAY_SIZE(SP0A28_preview_camera_regs), "preview_camera_regs");
    if (err) {
        return err;
    }
 #endif
 
    return 0;
}



static int camdrv_SP0A28_return_to_preview(struct v4l2_subdev *sd)
{
    
    struct i2c_client *client = v4l2_get_subdevdata(sd);
    int err = 0;

#if 0        
    // set SP0A28 to preview mode 
    err = camdrv_i2c_set_config_register(client, SP0A28_preview_camera_regs, ARRAY_SIZE(SP0A28_preview_camera_regs), "preview_camera_regs");
    if (err) {
        return err;
    }
 #endif
 
    return 0;

}


static int camdrv_SP0A28_set_capture_size(struct v4l2_subdev *sd)
{
	struct i2c_client *client = v4l2_get_subdevdata(sd);
	struct camdrv_state *state = to_state(sd);
	int err = 0, index;

	index = state->capture_framesize_index;

	CAM_INFO_MSG("index =%d \n", index);

	switch (index) {
	/* ======================= 1.333 Ratio ======================================= */
	case CAPTURE_SIZE_VGA: /* 416x320 */
	{
		if (ARRAY_SIZE(SP0A28_capture_size_640x480_regs) == 0) {
			CAM_ERROR_MSG("capture_size_640x480_regs not supported !!! \n");
		} else
			err = camdrv_i2c_set_config_register(client, SP0A28_capture_size_640x480_regs, ARRAY_SIZE(SP0A28_capture_size_640x480_regs), "capture_size_640x480_regs");

		state->postview_info.width = 640;
		state->postview_info.height = 480;
		break;
	}

	case CAPTURE_SIZE_WVGA:
	{
		if (ARRAY_SIZE(SP0A28_capture_size_800x480_regs) == 0) {
			CAM_ERROR_MSG("capture_size_800x480_regs not supported !!! \n");
		} else
			err = camdrv_i2c_set_config_register(client, SP0A28_capture_size_800x480_regs, ARRAY_SIZE(SP0A28_capture_size_800x480_regs), "capture_size_800x480_regs");

		state->postview_info.width = 800;
		state->postview_info.height = 480;
		break;
	}

	case CAPTURE_SIZE_SVGA:	/* 800x600 */
	{
		if (ARRAY_SIZE(SP0A28_capture_size_800x600_regs) == 0) {
			CAM_ERROR_MSG("capture_size_800x600_regs not supported !!! \n");
		} else
			err = camdrv_i2c_set_config_register(client, SP0A28_capture_size_800x600_regs, ARRAY_SIZE(SP0A28_capture_size_800x600_regs), "capture_size_800x600_regs");

		state->postview_info.width = 800;
		state->postview_info.height = 600;
		break;
	}

	case CAPTURE_SIZE_D1:		/* 720x480 */
	{
		if (ARRAY_SIZE(SP0A28_capture_size_720x480_regs) == 0) {
			CAM_ERROR_MSG("capture_size_720x480_regs not supported !!! \n");
		} else
			err = camdrv_i2c_set_config_register(client, SP0A28_capture_size_720x480_regs, ARRAY_SIZE(SP0A28_capture_size_720x480_regs), "capture_size_720x480_regs");

		state->postview_info.width = 720;
		state->postview_info.height = 480;
		break;
	}

	case CAPTURE_SIZE_WSVGA: 	/* 1024x600 */
	{
		if (ARRAY_SIZE(SP0A28_capture_size_1024x600_regs) == 0) {
			CAM_ERROR_MSG("capture_size_1024x600_regs not supported !!! \n");
		} else
			err = camdrv_i2c_set_config_register(client, SP0A28_capture_size_1024x600_regs, ARRAY_SIZE(SP0A28_capture_size_1024x600_regs), "capture_size_1024x600_regs");

		state->postview_info.width = 1024;
		state->postview_info.height = 600;
		break;
	}

	case CAPTURE_SIZE_XGA:	/* 1024x768 */
	{
		if (ARRAY_SIZE(SP0A28_capture_size_1024x768_regs) == 0) {
			CAM_ERROR_MSG("capture_size_1024x768_regs not supported !!! \n");
		} else
			err = camdrv_i2c_set_config_register(client, SP0A28_capture_size_1024x768_regs, ARRAY_SIZE(SP0A28_capture_size_1024x768_regs), "capture_size_1024x768_regs");

		state->postview_info.width = 1024;
		state->postview_info.height = 768;
		break;
	}

	case CAPTURE_SIZE_1MP:		/* 1280x960 */
	{
		if (ARRAY_SIZE(SP0A28_capture_size_1280x960_regs) == 0) {
			CAM_ERROR_MSG("capture_size_1280x960_regs not supported !!! \n");
		} else
			err = camdrv_i2c_set_config_register(client, SP0A28_capture_size_1280x960_regs, ARRAY_SIZE(SP0A28_capture_size_1280x960_regs), "capture_size_1280x960_regs");

		state->postview_info.width = 1024;
		state->postview_info.height = 960;
		break;
	}

	case CAPTURE_SIZE_W1MP:		/* 1600x960 */
	{
		if (ARRAY_SIZE(SP0A28_capture_size_1600x960_regs) == 0) {
			CAM_ERROR_MSG("capture_size_1600x960_regs not supported !!! \n");
		} else
			err = camdrv_i2c_set_config_register(client, SP0A28_capture_size_1600x960_regs, ARRAY_SIZE(SP0A28_capture_size_1600x960_regs), "capture_size_1600x960_regs");

		state->postview_info.width = 1600;
		state->postview_info.height = 960;
		break;
	}

	case CAPTURE_SIZE_2MP:	/* UXGA  - 1600x1200 */
	{
		if (ARRAY_SIZE(SP0A28_capture_size_1600x1200_regs) == 0) {
			CAM_ERROR_MSG("capture_size_1600x1200_regs not supported !!! \n");
		} else
			err = camdrv_i2c_set_config_register(client, SP0A28_capture_size_1600x1200_regs, ARRAY_SIZE(SP0A28_capture_size_1600x1200_regs), "capture_size_1600x1200_regs");

		state->postview_info.width = 1600;
		state->postview_info.height = 1200;
		break;
	}

	case CAPTURE_SIZE_W2MP:		/* 2048x1232, 2.4MP */
	{
		if (ARRAY_SIZE(SP0A28_capture_size_2048x1232_regs) == 0) {
			CAM_ERROR_MSG("capture_size_2048x1232_regs not supported !!! \n");
		} else
			err = camdrv_i2c_set_config_register(client, SP0A28_capture_size_2048x1232_regs, ARRAY_SIZE(SP0A28_capture_size_2048x1232_regs), "capture_size_2048x1232_regs");

		state->postview_info.width = 2048;
		state->postview_info.height = 1232;
		break;
	}

	case CAPTURE_SIZE_3MP:	/*2048x1536*/
	{
		if (ARRAY_SIZE(SP0A28_capture_size_2048x1536_regs) == 0) {
			CAM_ERROR_MSG("capture_size_2048x1536_regs not supported !!! \n");
		} else
			err = camdrv_i2c_set_config_register(client, SP0A28_capture_size_2048x1536_regs, ARRAY_SIZE(SP0A28_capture_size_2048x1536_regs), "capture_size_2048x1536_regs");

		state->postview_info.width = 2048;
		state->postview_info.height = 1536;
		break;
	}

	case CAPTURE_SIZE_5MP:		/* 2560x1920 */
	{
		if (ARRAY_SIZE(SP0A28_capture_size_2560x1920_regs) == 0) {
			CAM_ERROR_MSG("capture_size_2560x1920_regs not supported !!! \n");
		} else
			err = camdrv_i2c_set_config_register(client, SP0A28_capture_size_2560x1920_regs, ARRAY_SIZE(SP0A28_capture_size_2560x1920_regs), "capture_size_2560x1920_regs");

		state->postview_info.width = 2560;
		state->postview_info.height = 1920;
		break;
	}

	default:
	{
		CAM_ERROR_MSG("default case not supported !!! \n");
		err = -EINVAL;
		break;
	}
	}

	return err;
}




static int camdrv_SP0A28_capture_setting(struct v4l2_subdev *sd)
{
    // set SP0A28 to capture mode
    struct i2c_client *client = v4l2_get_subdevdata(sd);
    int err = 0;
    
    // Write capture setting
    err = camdrv_i2c_set_config_register(client, SP0A28_snapshot_normal_regs, ARRAY_SIZE(SP0A28_snapshot_normal_regs), "capture_camera_regs");
    if (err) {
        CAM_ERROR_MSG("Write capture setting was error !!!\n");
        return err;
    }

    // set the capture size
    err = camdrv_SP0A28_set_capture_size(sd);
    if (err) {
        CAM_ERROR_MSG("set the capture size was error !!!\n");
        return err;
    }
    
    return err;

}



#if 0
static int SP0A28_sensor_power(int on)
{
    CAM_INFO_MSG("[zhf]SP0A28_sensor_power is empty\n");

    return 0;
}

#else
/* @HW */
/* Power (common) */
static struct regulator *VCAM_IO_1_8_V = NULL;  //LDO_HV9
static struct regulator *VCAM_A_2_8_V = NULL;   //LDO_CAM12/12/2011

#define SENSOR_0_GPIO_PWRDN		(121)
#define SENSOR_0_GPIO_RST		(111)

#define SENSOR_1_GPIO_PWRDN		(122)
//#define SENSOR_1_CLK			"dig_ch1_clk"
#define SENSOR_1_CLK			"dig_ch0_clk"
#define SENSOR_1_CLK_FREQ		(26000000)


//#define SP0A28_ID 0x2284
#define SP0A28_ID 0xA2
static bool SP0A28_check_id(struct i2c_client *client)
{
    //u16 sensor_id = 0;
     u8 sensor_id = 0;
	int i;
/*
    camdrv_i2c_read_4_bytes(client, 0x0000, &sensor_id);
    CAM_INFO_MSG("+SP0A28_check_id = 0x%x \n", sensor_id);

    if(SP0A28_ID == sensor_id)
        return true;
    else
        return false;
*/
#if 0
	//for(i = 0; i<5; i++)
	while(1)
		{
			camdrv_i2c_write_2_bytes(client, 0xfd, 0x00);
			
			camdrv_i2c_read_2_bytes(client, 0x02, &sensor_id);
			 if(SP0A28_ID == sensor_id)
			 	break;
			 else
			 	{
			 	msleep(5);
				continue;

			 }
		}
    CAM_INFO_MSG("+SP0A28_check_id = 0x%x    ****************************\n", sensor_id);
	printk(KERN_ERR "+SP0A28_check_id = 0x%x    ****************************\n", sensor_id);

    if(SP0A28_ID == sensor_id)
        return true;
    else
        return false;
#endif
	return true;  //modify by qiuzhoujun 20120924
}


static int SP0A28_sensor_power(struct i2c_client *client, int on)
{
    unsigned int value;
    int ret = -1;
    struct clk *clock;
    struct clk *axi_clk;
    static struct pi_mgr_dfs_node unicam_dfs_node;

    CAM_INFO_MSG("camera power %s\n", (on ? "on" : "off"));

    if (!unicam_dfs_node.valid) {
        ret =
            pi_mgr_dfs_add_request(&unicam_dfs_node,"unicam", PI_MGR_PI_ID_MM,
                       PI_MGR_DFS_MIN_VALUE);
        if (ret) {
            printk(KERN_ERR
                   "%s: failed to register PI DFS request\n",
                   __func__);
            return -1;
        }
    }

    /*add by qiuzhoujun 20120925*/
	 if (gpio_request_one
        (SENSOR_0_GPIO_RST, GPIOF_DIR_OUT | GPIOF_INIT_LOW,
         "Cam0Rst")) {
        printk(KERN_ERR "%s: failed to get (Cam0Rst) gpio\n",
               __func__);
        return -1;
    }
    /*add end*/
	
    if (gpio_request_one(SENSOR_0_GPIO_PWRDN, GPIOF_DIR_OUT |
                 GPIOF_INIT_HIGH, "Cam0Pwr")) {
        printk(KERN_ERR "%s: failed to get (Cam0Pwr) gpio\n",
               __func__);
        return -1;
    }

	
    if (gpio_request_one(SENSOR_1_GPIO_PWRDN, GPIOF_DIR_OUT |
                 GPIOF_INIT_HIGH, "Cam1Pwr")) {
        printk(KERN_ERR "%s: failed to get (Cam1Pwr) gpio\n",
               __func__);
        return -1;
    }
    /*
    if (gpio_request_one(SENSOR_1_GPIO_PWRDN, GPIOF_DIR_OUT |
                 GPIOF_INIT_LOW, "Cam1Pwr")) {
        printk(KERN_ERR "%s: failed to get (Cam1Pwr) gpio\n",
               __func__);
        return -1;
    }
*/

    if(!VCAM_IO_1_8_V){
        VCAM_IO_1_8_V = regulator_get(NULL,"hv9");
        if(IS_ERR(VCAM_IO_1_8_V))
        {
            printk("can not get VCAM_IO_1.8V\n");
            return -1;
        }   
        regulator_set_voltage(VCAM_IO_1_8_V,1800000,1800000);
    }

    if(!VCAM_A_2_8_V){
        VCAM_A_2_8_V = regulator_get(NULL,"cam");
        if(IS_ERR(VCAM_A_2_8_V))
        {
            printk("can not get VCAM_A_2_8_V.8V\n");
            return -1;
        }
        regulator_set_voltage(VCAM_A_2_8_V,2800000,2800000);
    }


    clock = clk_get(NULL, SENSOR_1_CLK);
    if (IS_ERR_OR_NULL(clock)) {
        printk(KERN_ERR "%s: unable to get clock %s\n", __func__,
               SENSOR_1_CLK);
        return -1;
    }

    axi_clk = clk_get(NULL, "csi0_axi_clk");
    if (IS_ERR_OR_NULL(axi_clk)) {
        printk(KERN_ERR "%s:unable to get clock csi0_axi_clk\n",
               __func__);
        return -1;
    }

    if (on) {

        if (pi_mgr_dfs_request_update(&unicam_dfs_node, PI_OPP_TURBO)) {
            printk(KERN_ERR
                   "%s:failed to update dfs request for unicam\n",
                   __func__);
            return -1;
        }

        /* ------------Power on step1------------- */
	/*add by qiuzhoujun 20120925*/
	 gpio_set_value(SENSOR_0_GPIO_RST, 0);
	msleep(5);
	//gpio_set_value(SENSOR_0_GPIO_PWRDN, 1); 
	//msleep(10); 
	/*add end*/
        gpio_set_value(SENSOR_0_GPIO_PWRDN, 1);
        msleep(1);
        //gpio_set_value(SENSOR_1_GPIO_PWRDN, 1);
        gpio_set_value(SENSOR_1_GPIO_PWRDN, 0);  //modify by qiuzhoujun 20120925
        msleep(1);

        /* --------------Power on step2------------ */
        if(VCAM_IO_1_8_V){
            regulator_enable(VCAM_IO_1_8_V);
        }
        
        if(VCAM_A_2_8_V){
            regulator_enable(VCAM_A_2_8_V);
        }
       
        msleep(2);

        /* --------------Power on step3------------ */
        /* enable clk */
        value = clk_enable(axi_clk);
        if (value) {
            printk(KERN_ERR "%s:failed to enable csi2 axi clock\n",
                   __func__);
            return -1;
        }

        value = clk_enable(clock);
        if (value) {
            printk(KERN_ERR "%s: failed to enabled clock %s\n",
                   __func__, SENSOR_1_CLK);
            return -1;
        }
        value = clk_set_rate(clock, SENSOR_1_CLK_FREQ);
        if (value) {
            printk(KERN_ERR
                   "%s: failed to set the clock %s to freq %d\n",
                   __func__, SENSOR_1_CLK, SENSOR_1_CLK_FREQ);
            return -1;
        }
        msleep(2);
        
        /* --------------Power on step4------------ */
	gpio_set_value(SENSOR_1_GPIO_PWRDN, 0);
	msleep(100);
	gpio_set_value(SENSOR_1_GPIO_PWRDN, 1);  //add by qiuzhoujun 20120925
	msleep(100); //add by qiuzhoujun 20120925
		
        gpio_set_value(SENSOR_1_GPIO_PWRDN, 0);
        //msleep(30);
        msleep(100);  //modify by qiuzhoujun 20120925

        if(!SP0A28_check_id(client)){
            gpio_free(SENSOR_0_GPIO_RST);
            gpio_free(SENSOR_0_GPIO_PWRDN);
            gpio_free(SENSOR_1_GPIO_PWRDN);
            return -1;
        }

    } else {

        /* --------------Power off step1------------ */
        gpio_set_value(SENSOR_1_GPIO_PWRDN, 1);
        msleep(5);
		
        /* --------------Power off step2------------ */
        clk_disable(clock);
        clk_disable(axi_clk);
        msleep(5);

        /* --------------Power off step3------------ */
        if(VCAM_A_2_8_V){
            regulator_disable(VCAM_A_2_8_V);
        }
        if(VCAM_IO_1_8_V){
            regulator_disable(VCAM_IO_1_8_V);
        }
        msleep(5);

        /* --------------Power off step4------------ */
        /* enable power down gpio */
        gpio_set_value(SENSOR_0_GPIO_PWRDN, 0);
        msleep(1);
        gpio_set_value(SENSOR_1_GPIO_PWRDN, 0);
        msleep(1);

        if (pi_mgr_dfs_request_update(&unicam_dfs_node,
                          PI_MGR_DFS_MIN_VALUE)) {
            printk(KERN_ERR "%s: failed to update dfs request for unicam\n",
                 __func__);
        }
    }

    gpio_free(SENSOR_0_GPIO_PWRDN);
    gpio_free(SENSOR_1_GPIO_PWRDN);
     gpio_free(SENSOR_0_GPIO_RST);  //add by qiuzhoujun 20120925
	printk(KERN_ERR "in the end of SP0A28_sensor_power ***********\n");
    return 0;
}


#endif

static bool camdrv_sensor_SP0A28_init(bool bOn, struct camdrv_sensor_cap *sensor)
{
	//camdrv_SP0A28_power(bOn);
	strcpy(sensor->name, SP0A28_NAME);

	sensor->supported_preview_framesize_list
		= SP0A28_supported_preview_framesize_list;
	sensor->supported_number_of_preview_sizes
		= ARRAY_SIZE(SP0A28_supported_preview_framesize_list);

	sensor->supported_capture_framesize_list
		= SP0A28_supported_capture_framesize_list;
	sensor->supported_number_of_capture_sizes
		= ARRAY_SIZE(SP0A28_supported_capture_framesize_list);

	sensor->fmts
		= SP0A28_fmts;
	sensor->rows_num_fmts
		= ARRAY_SIZE(SP0A28_fmts);

	sensor->controls
		= SP0A28_controls;
	sensor->rows_num_controls
		= ARRAY_SIZE(SP0A28_controls);

	sensor->default_pix_fmt
		= SP0A28_DEFAULT_PIX_FMT;
	sensor->default_mbus_pix_fmt
		= SP0A28_DEFAULT_MBUS_PIX_FMT;
	sensor->register_size
		= SP0A28_REGISTER_SIZE;


    //zhuohengfeng++ for turnkey camera compatible
    sensor->MIPI_serial.lanes = SP0A28_MIPI_LANES;
    sensor->MIPI_serial.channel = SP0A28_MIPI_CHANNEL;
    sensor->MIPI_serial.hs_term_time = SP0A28_MIPI_HS_TERM_TIME;
    sensor->MIPI_serial.hs_cap_term_time = SP0A28_MIPI_HS_TERM_TIME;


/* sensor dependent functions */
	/* mandatory */
	sensor->thumbnail_ioctl
		= camdrv_SP0A28_ss_ioctl;
	sensor->enum_frameintervals
		= camdrv_SP0A28_enum_frameintervals;

	/* optional */

    sensor->preview_setting
        = camdrv_SP0A28_preview_setting;
    sensor->return_to_preview
        = camdrv_SP0A28_return_to_preview;
    sensor->capture_setting
        = camdrv_SP0A28_capture_setting;
	sensor->i2c_set_data_burst
		= camdrv_SP0A28_i2c_set_data_burst;

	/* REGS and their sizes */
	/* List all the capabilities of sensor . List all the supported register setting tables */

	sensor->init_regs
		= SP0A28_init_regs;
	sensor->rows_num_init_regs
		= ARRAY_SIZE(SP0A28_init_regs);

	//sensor->preview_camera_regs
	//	= SP0A28_preview_camera_regs;
	//sensor->rows_num_preview_camera_regs
	//	= ARRAY_SIZE(SP0A28_preview_camera_regs);


	/* snapshot mode */
	sensor->snapshot_normal_regs
		= SP0A28_snapshot_normal_regs;
	sensor->rows_num_snapshot_normal_regs
		= ARRAY_SIZE(SP0A28_snapshot_normal_regs);

	sensor->snapshot_lowlight_regs
		= SP0A28_snapshot_lowlight_regs;
	sensor->rows_num_snapshot_lowlight_regs
		= ARRAY_SIZE(SP0A28_snapshot_lowlight_regs);

	sensor->snapshot_highlight_regs
		= SP0A28_snapshot_highlight_regs;
	sensor->rows_num_snapshot_highlight_regs
		= ARRAY_SIZE(SP0A28_snapshot_highlight_regs);

	sensor->snapshot_nightmode_regs
		= SP0A28_snapshot_nightmode_regs;
	sensor->rows_num_snapshot_nightmode_regs
		= ARRAY_SIZE(SP0A28_snapshot_nightmode_regs);



	/* effect */
	sensor->effect_normal_regs
		= SP0A28_effect_normal_regs;
	sensor->rows_num_effect_normal_regs
		= ARRAY_SIZE(SP0A28_effect_normal_regs);

	sensor->effect_negative_regs
		= SP0A28_effect_negative_regs;
	sensor->rows_num_effect_negative_regs
		= ARRAY_SIZE(SP0A28_effect_negative_regs);

	sensor->effect_sepia_regs
		= SP0A28_effect_sepia_regs;
	sensor->rows_num_effect_sepia_regs
		= ARRAY_SIZE(SP0A28_effect_sepia_regs);

	sensor->effect_mono_regs
		= SP0A28_effect_mono_regs;
	sensor->rows_num_effect_mono_regs
		= ARRAY_SIZE(SP0A28_effect_mono_regs);

	sensor->effect_aqua_regs
		= SP0A28_effect_aqua_regs;
	sensor->rows_num_effect_aqua_regs
		= ARRAY_SIZE(SP0A28_effect_aqua_regs);

	sensor->effect_sharpen_regs
		= SP0A28_effect_sharpen_regs;
	sensor->rows_num_effect_sharpen_regs
		= ARRAY_SIZE(SP0A28_effect_sharpen_regs);

	sensor->effect_solarization_regs
		= SP0A28_effect_solarization_regs;
	sensor->rows_num_effect_solarization_regs
		= ARRAY_SIZE(SP0A28_effect_solarization_regs);

	sensor->effect_black_white_regs
		= SP0A28_effect_black_white_regs;
	sensor->rows_num_effect_black_white_regs
		= ARRAY_SIZE(SP0A28_effect_black_white_regs);

	/* wb */
	sensor->wb_auto_regs
		= SP0A28_wb_auto_regs;
	sensor->rows_num_wb_auto_regs
		= ARRAY_SIZE(SP0A28_wb_auto_regs);

	sensor->wb_sunny_regs
		= SP0A28_wb_sunny_regs;
	sensor->rows_num_wb_sunny_regs
		= ARRAY_SIZE(SP0A28_wb_sunny_regs);

	sensor->wb_cloudy_regs
		= SP0A28_wb_cloudy_regs;
	sensor->rows_num_wb_cloudy_regs
		= ARRAY_SIZE(SP0A28_wb_cloudy_regs);

	sensor->wb_tungsten_regs
		= SP0A28_wb_tungsten_regs;
	sensor->rows_num_wb_tungsten_regs
		= ARRAY_SIZE(SP0A28_wb_tungsten_regs);

	sensor->wb_fluorescent_regs
		= SP0A28_wb_fluorescent_regs;
	sensor->rows_num_wb_fluorescent_regs
		= ARRAY_SIZE(SP0A28_wb_fluorescent_regs);

	sensor->wb_cwf_regs
		= SP0A28_wb_cwf_regs;
	sensor->rows_num_wb_cwf_regs
		= ARRAY_SIZE(SP0A28_wb_cwf_regs);


	/* EV */
	sensor->ev_minus_4_regs
		= SP0A28_ev_minus_4_regs;
	sensor->rows_num_ev_minus_4_regs
		= ARRAY_SIZE(SP0A28_ev_minus_4_regs);

	sensor->ev_minus_3_regs
		= SP0A28_ev_minus_3_regs;
	sensor->rows_num_ev_minus_3_regs
		= ARRAY_SIZE(SP0A28_ev_minus_3_regs);

	sensor->ev_minus_2_regs
		= SP0A28_ev_minus_2_regs;
	sensor->rows_num_ev_minus_2_regs
		= ARRAY_SIZE(SP0A28_ev_minus_2_regs);

	sensor->ev_minus_1_regs
		= SP0A28_ev_minus_1_regs;
	sensor->rows_num_ev_minus_1_regs
		= ARRAY_SIZE(SP0A28_ev_minus_1_regs);

	sensor->ev_default_regs
		= SP0A28_ev_default_regs;
	sensor->rows_num_ev_default_regs
		= ARRAY_SIZE(SP0A28_ev_default_regs);

	sensor->ev_plus_1_regs
		= SP0A28_ev_plus_1_regs;
	sensor->rows_num_ev_plus_1_regs
		= ARRAY_SIZE(SP0A28_ev_plus_1_regs);

	sensor->ev_plus_2_regs
		= SP0A28_ev_plus_2_regs;
	sensor->rows_num_ev_plus_2_regs
		= ARRAY_SIZE(SP0A28_ev_plus_2_regs);

	sensor->ev_plus_3_regs
		= SP0A28_ev_plus_3_regs;
	sensor->rows_num_ev_plus_3_regs
		= ARRAY_SIZE(SP0A28_ev_plus_3_regs);

	sensor->ev_plus_4_regs
		= SP0A28_ev_plus_4_regs;
	sensor->rows_num_ev_plus_4_regs
		= ARRAY_SIZE(SP0A28_ev_plus_4_regs);


	/* preview size */
	sensor->preview_size_176x144_regs
		= SP0A28_preview_size_176x144_regs;
	sensor->rows_num_preview_size_176x144_regs
		= ARRAY_SIZE(SP0A28_preview_size_176x144_regs);

	sensor->preview_size_320x240_regs
		= SP0A28_preview_size_320x240_regs;
	sensor->rows_num_preview_size_320x240_regs
		= ARRAY_SIZE(SP0A28_preview_size_320x240_regs);

	sensor->preview_size_352x288_regs
		= SP0A28_preview_size_352x288_regs;
	sensor->rows_num_preview_size_352x288_regs
		= ARRAY_SIZE(SP0A28_preview_size_352x288_regs);

	sensor->preview_size_640x480_regs
		= SP0A28_preview_size_640x480_regs;
	sensor->rows_num_preview_size_640x480_regs
		= ARRAY_SIZE(SP0A28_preview_size_640x480_regs);

	sensor->preview_size_704x576_regs
		= SP0A28_preview_size_704x576_regs;
	sensor->rows_num_preview_size_704x576_regs
		= ARRAY_SIZE(SP0A28_preview_size_704x576_regs);

	sensor->preview_size_720x480_regs
		= SP0A28_preview_size_720x480_regs;
	sensor->rows_num_preview_size_720x480_regs
		= ARRAY_SIZE(SP0A28_preview_size_720x480_regs);

	sensor->preview_size_800x480_regs
		= SP0A28_preview_size_800x480_regs;
	sensor->rows_num_preview_size_800x480_regs
		= ARRAY_SIZE(SP0A28_preview_size_800x480_regs);

	sensor->preview_size_800x600_regs
		= SP0A28_preview_size_800x600_regs;
	sensor->rows_num_preview_size_800x600_regs
		= ARRAY_SIZE(SP0A28_preview_size_800x600_regs);

	sensor->preview_size_1024x600_regs
		= SP0A28_preview_size_1024x600_regs;
	sensor->rows_num_preview_size_1024x600_regs
		= ARRAY_SIZE(SP0A28_preview_size_1024x600_regs);

	sensor->preview_size_1024x768_regs
		= SP0A28_preview_size_1024x768_regs;
	sensor->rows_num_preview_size_1024x768_regs
		= ARRAY_SIZE(SP0A28_preview_size_1024x768_regs);

	sensor->preview_size_1280x960_regs
		= SP0A28_preview_size_1280x960_regs;
	sensor->rows_num_preview_size_1280x960_regs
		= ARRAY_SIZE(SP0A28_preview_size_1280x960_regs);

	sensor->preview_size_1600x960_regs
		= SP0A28_preview_size_1600x960_regs;
	sensor->rows_num_preview_size_1600x960_regs
		= ARRAY_SIZE(SP0A28_preview_size_1600x960_regs);

	sensor->preview_size_1600x1200_regs
		= SP0A28_preview_size_1600x1200_regs;
	sensor->rows_num_preview_size_1600x1200_regs
		= ARRAY_SIZE(SP0A28_preview_size_1600x1200_regs);

	sensor->preview_size_2048x1232_regs
		= SP0A28_preview_size_2048x1232_regs;
	sensor->rows_num_preview_size_2048x1232_regs
		= ARRAY_SIZE(SP0A28_preview_size_2048x1232_regs);

	sensor->preview_size_2048x1536_regs
		= SP0A28_preview_size_2048x1536_regs;
	sensor->rows_num_preview_size_2048x1536_regs
		= ARRAY_SIZE(SP0A28_preview_size_2048x1536_regs);

	sensor->preview_size_2560x1920_regs
		= SP0A28_preview_size_2560x1920_regs;
	sensor->rows_num_preview_size_2560x1920_regs
		= ARRAY_SIZE(SP0A28_preview_size_2560x1920_regs);



	sensor->sensor_power
		= SP0A28_sensor_power;
    
    return true;
};

const struct camdrv_sensor_cfg SP0A28_cfg = {
    .i2c_address = (0x7A>>1),
    .camdrv_sensor_init  = camdrv_sensor_SP0A28_init,
};
EXPORT_SYMBOL(SP0A28_cfg);
