/* drivers/media/video/camdrv_sp0a28.h
 *
 * Driver for sp0a28 (5MP Camera) from SEC(LSI), firmware EVT1.1
 *
 * Copyright (C) 2012, OV
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */


#ifndef __CAMDRV_SP0A28_H__
#define __CAMDRV_SP0A28_H__

#define START_BURST_MODE 0x0F12

#define SP0A28_SATURATION_MIN			0
#define SP0A28_SATURATION_MAX			200
#define SP0A28_SATURATION_STEP          100
#define SP0A28_SATURATION_DEF			100

#define SP0A28_SHARPNESS_MIN			0
#define SP0A28_SHARPNESS_MAX			200
#define SP0A28_SHARPNESS_STEP           100
#define SP0A28_SHARPNESS_DEF			100

/* Auto Focus Command */
#define SP0A28_CMD_MAIN					0x3022
#define SP0A28_CMD_ACK					0x3023
#define SP0A28_CMD_FW_STATUS				0x3029
/*  Auto Focus Status */
#define SP0A28_AF_NORMALIZED_W				80
#define SP0A28_AF_NORMALIZED_H				60
#define SP0A28_AF_SUCCESS				0
#define SP0A28_AF_PENDING				1
#define SP0A28_AF_FAIL					2

#define SP0A28_NOT_FOCUSING				0
#define SP0A28_FOCUSING					1

#define SP0A28_MAX_FOCUS_AREAS 5

static const struct sensor_reg sp0a8_init_regs[] = {

   {8, 0xfd , 0x01},
  {8, 0x7c , 0x6c},
  {8, 0xfd , 0x00},  
  {8, 0x1C , 0x00},
  {8, 0x32 , 0x00},
  {8, 0x0e , 0x00},
  {8, 0x0f , 0x40},
  {8, 0x10 , 0x40},
  {8, 0x11 , 0x10},
  {8, 0x12 , 0xa0},
  {8, 0x13 , 0x40},
  {8, 0x15 , 0x00},
  {8, 0x16 , 0x08},
  {8, 0x1A , 0x37},
  {8, 0x1B , 0x17},
  {8, 0x1C , 0x2f},
  {8, 0x1d , 0x00},
  {8, 0x1E , 0x57},
  {8, 0x21 , 0x36},   //2f
  {8, 0x22 , 0x12},
  {8, 0x24 , 0x80},
  {8, 0x25 , 0x02},
  {8, 0x26 , 0x03},
  {8, 0x27 , 0xeb},
  {8, 0x28 , 0x5f},
  {8, 0x2f , 0x01},
  {8, 0x5f , 0x02},
  {8, 0xfb , 0x33}, 
  {8, 0xe7 , 0x03}, 
  {8, 0xe7 , 0x00},
//;ae setting
  {8, 0xfd, 0x00},
  {8, 0x03, 0x01},
  {8, 0x04, 0x2c},
  {8, 0x05, 0x00},
  {8, 0x06, 0x00},
  {8, 0x09, 0x02},
  {8, 0x0a, 0x80},
  {8, 0xf0, 0x64},
  {8, 0xf1, 0x00},
  {8, 0xfd, 0x01},
  {8, 0x90, 0x0c},
  {8, 0x92, 0x01},
  {8, 0x98, 0x64},
  {8, 0x99, 0x00},
  {8, 0x9a, 0x02},
  {8, 0x9b, 0x00},
//;Status
  {8, 0xfd, 0x01},
  {8, 0xce, 0xb0},
  {8, 0xcf, 0x04},
  {8, 0xd0, 0xb0},
  {8, 0xd1, 0x04},
  {8, 0xfd, 0x00},
  
  {8, 0xfd, 0x01},
  {8, 0xc4, 0x6c},
  {8, 0xc5, 0x7c},
  {8, 0xca, 0x30},
  {8, 0xcb, 0x45},
  {8, 0xcc, 0x60},
  {8, 0xcd, 0x60},
  
//;lsc  for SX5044module
  {8, 0xfd, 0x01},
  {8, 0x35, 0x0e},
  {8, 0x36, 0x14},
  {8, 0x37, 0x20},
  {8, 0x38, 0x1c},
  {8, 0x39, 0x09},
  {8, 0x3a, 0x10},
  {8, 0x3b, 0x19},
  {8, 0x3c, 0x1a},
  {8, 0x3d, 0x08},
  {8, 0x3e, 0x10},
  {8, 0x3f, 0x14},
  {8, 0x40, 0x22},
  {8, 0x41, 0x0a},
  {8, 0x42, 0x00},
  {8, 0x43, 0x0a},
  {8, 0x44, 0xfc},
  {8, 0x45, 0x00},   //;00;3
  {8, 0x46, 0xfe},
  {8, 0x47, 0x00},
  {8, 0x48, 0xf8},
  {8, 0x49, 0x00},
  {8, 0x4a, 0xf8},
  {8, 0x4b, 0x00},
  {8, 0x4c, 0xf2},
  {8, 0xfd, 0x00},
  {8, 0xa1, 0x20},
  {8, 0xa2, 0x20},
  {8, 0xa3, 0x20},
  {8, 0xa4, 0xff},        
   
// ;smooth
  {8, 0xfd,0x01},
  {8, 0xde,0x0f},
 
  {8, 0xfd, 0x00},
	 //ͨƽֵ	
  {8, 0x57, 0x04}, //	;raw_dif_thr_outdoor
  {8, 0x58, 0x0a}, //;raw_dif_thr_normal
  {8, 0x56, 0x10}, //;raw_dif_thr_dummy
  {8, 0x59, 0x10}, //;raw_dif_thr_lowlight
	 //GrGbƽֵ
  {8, 0x89, 0x04}, //	;raw_grgb_thr_outdoor 
  {8, 0x8a, 0x0a}, // ;raw_grgb_thr_normal  
  {8, 0x9c, 0x10}, // ;raw_grgb_thr_dummy   
  {8, 0x9d, 0x10}, // ;raw_grgb_thr_lowlight
	 //Gr\Gb֮ƽǿ
  {8, 0x81, 0xe0}, //    ;raw_gflt_fac_outdoor
  {8, 0x82, 0xa0}, // ;80;raw_gflt_fac_normal
  {8, 0x83, 0x80}, //    ;raw_gflt_fac_dummy
  {8, 0x84, 0x80}, //    ;raw_gflt_fac_lowlight
  //GrGbͨƽǿ  
  {8, 0x85, 0xe0},  // ;raw_gf_fac_outdoor  
  {8, 0x86, 0xa0},  // ;raw_gf_fac_normal  
  {8, 0x87, 0x80},  // ;raw_gf_fac_dummy   
  {8, 0x88, 0x80},  // ;raw_gf_fac_lowlight
  //RBƽǿ  
  {8, 0x5a, 0xff}, //		 ;raw_rb_fac_outdoor
  {8, 0x5b, 0xc0}, // ;40;raw_rb_fac_normal
  {8, 0x5c, 0xa0}, // 	 ;raw_rb_fac_dummy
  {8, 0x5d, 0xa0}, // 	 ;raw_rb_fac_lowlight
  // ƽֵӦ
  {8, 0xa7, 0xff}, //
  {8, 0xa8, 0xff}, //;0x2f
  {8, 0xa9, 0xff}, //;0x2f
  {8, 0xaa, 0xff}, //;0x2f

//;dem_morie_thr ȥĦ
  {8, 0x9e, 0x10},

//;sharpen 
  {8, 0xfd, 0x01}, //
  {8, 0xe2, 0x30}, //	;sharpen_y_base
  {8, 0xe4, 0xa0}, //	;sharpen_y_max               
  {8, 0xe5, 0x08}, // ;rangek_neg_outdoor
  {8, 0xd3, 0x10}, // ;rangek_pos_outdoor   
  {8, 0xd7, 0x08}, // ;range_base_outdoor                 
  {8, 0xe6, 0x08}, // ;rangek_neg_normal
  {8, 0xd4, 0x10}, // ;rangek_pos_normal 
  {8, 0xd8, 0x08}, // ;range_base_normal                
  {8, 0xe7, 0x10}, // ;rangek_neg_dummy
  {8, 0xd5, 0x10}, // ;rangek_pos_dummy
  {8, 0xd9, 0x10}, // ;range_base_dummy 	        
  {8, 0xd2, 0x10}, // ;rangek_neg_lowlight
  {8, 0xd6, 0x10}, // ;rangek_pos_lowlight
  {8, 0xda, 0x10}, // ;range_base_lowlight
  {8, 0xe8, 0x20}, //;sharp_fac_pos_outdoor
  {8, 0xec, 0x30}, //;sharp_fac_neg_outdoor
  {8, 0xe9, 0x10}, //;sharp_fac_pos_nr
  {8, 0xed, 0x30}, //;sharp_fac_neg_nr
  {8, 0xea, 0x10}, //;sharp_fac_pos_dummy
  {8, 0xef, 0x20}, //;sharp_fac_neg_dummy
  {8, 0xeb, 0x10}, //;sharp_fac_pos_low
  {8, 0xf0, 0x10}, //;sharp_fac_neg_low 
     
//;CCM
  {8, 0xfd, 0x01}, //
  {8, 0xa0, 0x80}, //;8c;80;80;80(ɫӽɫ)
  {8, 0xa1, 0x00}, //;0c;00;0 ;0 
  {8, 0xa2, 0x00}, //;e8;00;0 ;0 
  {8, 0xa3, 0xf6}, //;ec;ff;f2;f3;f0
  {8, 0xa4, 0x99}, //;99;9a;8e;a6
  {8, 0xa5, 0xf2}, //;fb;e7;0 ;ea
  {8, 0xa6, 0x0d}, //;0d;0c;0 ;0 
  {8, 0xa7, 0xda}, //;da;da;e6;e6
  {8, 0xa8, 0x98}, //;98;9a;9a;9a
  {8, 0xa9, 0x00}, //;30;00;0 ;0 
  {8, 0xaa, 0x33}, //;33;33;3 ;33
  {8, 0xab, 0x0c}, //;0c;0c;c ;c 
  {8, 0xfd, 0x00}, //;00	 

//;gamma  
  {8, 0xfd, 0x00}, //;00
  {8, 0x8b, 0x00}, //;00;0 ;0      
  {8, 0x8c, 0x0f}, //;11;C ;12
  {8, 0x8d, 0x21}, //;24;19;1f
  {8, 0x8e, 0x3b}, //;3f;2C;31
  {8, 0x8f, 0x64}, //;64;49;4c
  {8, 0x90, 0x84}, //;7f;61;62
  {8, 0x91, 0xa0}, //;93;77;77
  {8, 0x92, 0xb6}, //;a4;8A;89
  {8, 0x93, 0xc7}, //;b2;9B;9b
  {8, 0x94, 0xd2}, //;bb;A9;a8
  {8, 0x95, 0xda}, //;c4;B5;b5
  {8, 0x96, 0xe0}, //;cb;C0;c0
  {8, 0x97, 0xe6}, //;d1;CA;ca
  {8, 0x98, 0xea}, //;d5;D4;d4
  {8, 0x99, 0xef}, //;dc;DD;dd
  {8, 0x9a, 0xf4}, //;e2;E6;e6
  {8, 0x9b, 0xf7}, //;e9;EF;ef
  {8, 0xfd, 0x01}, //;01;01;01
  {8, 0x8d, 0xfb}, //;f2;F7;f7
  {8, 0x8e, 0xff}, //;ff;FF;ff
  {8, 0xfd, 0x00}, //;00;00;00

//; awb
  {8, 0xfd, 0x01},
  {8, 0x28, 0xc4},	
  {8, 0x29, 0x9e},
  {8, 0x11, 0x13},	
  {8, 0x12, 0x13},
  {8, 0x2e, 0x0d},	
  {8, 0x2f, 0x0d},
  {8, 0x16, 0x1c},
  {8, 0x17, 0x1a},
  {8, 0x18, 0x1a},	
  {8, 0x19, 0x54},	
  {8, 0x1a, 0xa5},  
  {8, 0x1b, 0x9a},
  {8, 0x2a, 0xef},
 
//;AE
//;rpc
  {8, 0xfd, 0x00},  //
  {8, 0xe0, 0x3a},  //;38;rpc_1base_max
  {8, 0xe1, 0x2c},  //;2a;rpc_2base_max
  {8, 0xe2, 0x26},  //;26;rpc_3base_max
  {8, 0xe3, 0x22},  //;rpc_4base_max
  {8, 0xe4, 0x22},  //;rpc_5base_max
  {8, 0xe5, 0x20},  //;rpc_6base_max
  {8, 0xe6, 0x20},  //;rpc_7base_max
  {8, 0xe8, 0x20},  //;1e;rpc_8base_max
  {8, 0xe9, 0x20},  //;1e;rpc_9base_max
  {8, 0xea, 0x20},  //;1e;rpc_10base_max
  {8, 0xeb, 0x1e},  //;1d;rpc_11base_max
  {8, 0xf5, 0x1e},  //;1d;rpc_12base_max
  {8, 0xf6, 0x1e},  //;1d;rpc_13base_max
//;ae min gain  
  {8, 0xfd, 0x01},  //
  {8, 0x94, 0x60},  //;rpc_max_indr
  {8, 0x95, 0x1e},  //;rpc_min_indr 
  {8, 0x9c, 0x60},  //;rpc_max_outdr
  {8, 0x9d, 0x1e},  //;rpc_min_outdr   


//;ae target,
  {8, 0xfd, 0x00}, // 
  {8, 0xed, 0x7c}, //;90 
  {8, 0xf7, 0x78}, //;8c 
  {8, 0xf8, 0x70}, //;84 
  {8, 0xec, 0x6c}, //;80  
  {8, 0xef, 0x7c}, //
  {8, 0xf9, 0x78}, //
  {8, 0xfa, 0x70}, //
  {8, 0xee, 0x6c}, //
  
//;gray detect
  {8, 0xfd, 0x01},
  {8, 0x30, 0x40},
  {8, 0x31, 0x70},
  {8, 0x32, 0x20},
  {8, 0x33, 0xef},
  {8, 0x34, 0x02},
  {8, 0x4d, 0x40},
  {8, 0x4e, 0x15},
  {8, 0x4f, 0x13},
       
//;saturation
  {8, 0xfd, 0x00}, //
  {8, 0xbe, 0xaa}, //  
  {8, 0xc0, 0xff}, //
  {8, 0xc1, 0xff}, //
  {8, 0xd3, 0x90}, //;78
  {8, 0xd4, 0x90}, //;78 
  {8, 0xd6, 0x70}, //
  {8, 0xd7, 0x60}, //
  {8, 0xd8, 0x90}, //;78
  {8, 0xd9, 0x90}, //;78 
  {8, 0xda, 0x70}, //
  {8, 0xdb, 0x60}, //

//;heq
  {8, 0xfd, 0x00},  //
  {8, 0xdc, 0x00},  //;heq_offset
  {8, 0xdd, 0x80},  //;ku
  {8, 0xde, 0x80},  //;90;kl
  {8, 0xdf, 0x80},  //;heq_mean
  
//;YCnr
  {8, 0xfd, 0x00},      //
  {8, 0xc2, 0x08},      //	;Ynr_thr_outdoor
  {8, 0xc3, 0x08},      //	;Ynr_thr_normal
  {8, 0xc4, 0x08},      //	;Ynr_thr_dummy
  {8, 0xc5, 0x10},      //	;Ynr_thr_lowlight
  {8, 0xc6, 0x80},      //	;cnr_thr_outdoor
  {8, 0xc7, 0x80},      //	;cnr_thr_normal  
  {8, 0xc8, 0x80},      //	;cnr_thr_dummy   
  {8, 0xc9, 0x80},      //	;cnr_thr_lowlight  

//;auto lum
  {8, 0xfd, 0x00},  
  {8, 0xb2, 0x18},  //;10	
  {8, 0xb3, 0x1f},   
  {8, 0xb4, 0x30},   
  {8, 0xb5, 0x45},   
      
//;func enable
  {8, 0xfd, 0x00},
  {8, 0x32, 0xd },
  {8, 0x34, 0x1e},
  {8, 0x33, 0xef},

};

static const struct sensor_reg sp0a8_preview_camera_regs[] = {

	//keep empty

};

static const struct sensor_reg sp0a8_snapshot_normal_regs[] = {

	//keep empty
};




/* ========================================================== */
/*   EFFECT(6) */
/* ========================================================== */

static const struct sensor_reg sp0a8_effect_normal_regs[] = 
{	//
	{0xfd, 0x00},
	{0x62, 0x00},
	{0x63, 0x80},
	{0x64, 0x80},
};

static const struct sensor_reg sp0a8_effect_negative_regs[] = {//
	{0xfd, 0x00},
	{0x62, 0x08},
	{0x63, 0x80},
	{0x64, 0x80},
};

static const struct sensor_reg sp0a8_effect_sepia_regs[] = {//
	{0xfd, 0x00},
	{0x62, 0x10},
	{0x63, 0xc0},
	{0x64, 0x20},
};
static const struct sensor_reg sp0a8_effect_mono_regs[] = {//
	{0xfd, 0x00},
	{0x62, 0x20},
	{0x63, 0x80},
	{0x64, 0x80},
};
static const struct sensor_reg sp0a8_effect_aqua_regs[] = {
	{0xfd, 0x00},
	{0x62, 0x10},
	{0x63, 0x60},
	{0x64, 0x60},
};
static const struct sensor_reg sp0a8_effect_sharpen_regs[] = {
	{0xfd, 0x00},
	{0x62, 0x10},
	{0x63, 0x20},
	{0x64, 0xd0},
};
static const struct sensor_reg sp0a8_effect_solarization_regs[] = {
//
};
static const struct sensor_reg sp0a8_effect_black_white_regs[] = {
//
};
/* ========================================================== */
/*   EFFECT(6)  -END */
/* ========================================================== */


/* ========================================================== */
/*  WB(5) -START */
/* ========================================================== */

static const struct sensor_reg sp0a8_wb_auto_regs[] = {
	{0xfd,0x01}													   
	{0x28,0xc4},															   
	{0x29,0x9e},
	{0xfd,0x00},   // AUTO 3000K~7000K		  
	{0x32,0x0d},	   
};

static const struct sensor_reg sp0a8_wb_sunny_regs[] = {
	{0xfd,0x00},	//6500K 									
	{0x32,0x05},															
	{0xfd,0x01},															
	{0x28,0xc1},																
	{0x29,0x88},																
	{0xfd,0x00},		
};

static const struct sensor_reg sp0a8_wb_cloudy_regs[] = {
	{0xfd,0x00},	 //7000K									 
	{0x32,0x05},															
	{0xfd,0x01},															
	{0x28,0xc1},																
	{0x29,0x88},																
	{0xfd,0x00},
};

static const struct sensor_reg sp0a8_wb_tungsten_regs[] = {
	{0xfd,0x00},   //4000K									 
	{0x32,0x05},														   
	{0xfd,0x01},														   
	{0x28,0xae},															   
	{0x29,0xcc},															   
	{0xfd,0x00},   
};

static const struct sensor_reg sp0a8_wb_fluorescent_regs[] = {
	{0xfd,0x00},   //4000K									 
	{0x32,0x05},														   
	{0xfd,0x01},														   
	{0x28,0xae},															   
	{0x29,0xcc},															   
	{0xfd,0x00},   
};


/* ========================================================== */
/*  WB(5)  --END */
/* ========================================================== */





/* ========================================================== */
/*  BRIGHTNES(9) - START */
/* ========================================================== */

static const struct sensor_reg sp0a8_ev_minus_4_regs[] = {
	{0xfd, 0x00},
	{0xdc, 0xc0},
};

static const struct sensor_reg sp0a8_ev_minus_3_regs[] = {
	{0xfd, 0x00},
	{0xdc, 0xd0},
};

static const struct sensor_reg sp0a8_ev_minus_2_regs[] = {
	{0xfd, 0x00},
	{0xdc, 0xe0},
};

static const struct sensor_reg sp0a8_ev_minus_1_regs[] = {
	{0xfd, 0x00},
	{0xdc, 0xf0},
};

static const struct sensor_reg sp0a8_ev_default_regs[] = {
	{0xfd, 0x00},
	{0xdc, 0x00},
};

static const struct sensor_reg sp0a8_ev_plus_1_regs[] = {
	{0xfd, 0x00},
	{0xdc, 0x10},
};

static const struct sensor_reg sp0a8_ev_plus_2_regs[] = {
	{0xfd, 0x00},
	{0xdc, 0x20},
};

static const struct sensor_reg sp0a8_ev_plus_3_regs[] = {
	{0xfd, 0x00},
	{0xdc, 0x30},
};

static const struct sensor_reg sp0a8_ev_plus_4_regs[] = {
	{0xfd, 0x00},
	{0xdc, 0x40},
};
/*  ========================================================== */
/*  BRIGHTNES(9) - END */
/* ========================================================== */

/* ========================================================== */
/*  contrast(5) - START */
/* ========================================================== */


static const struct sensor_reg sp0a8_contrast_minus_2_regs[] = 
{
	{0xfd,0x00},
	{0xdd,0x70},
	{0xde,0x60},
};

static const struct sensor_reg sp0a8_contrast_minus_1_regs[] = 
{
	{0xfd,0x00},
	{0xdd,0x70},
	{0xde,0x70},
};

static const struct sensor_reg sp0a8_contrast_default_regs[] = 
{
	{0xfd,0x00},
	{0xdd,0x80},
	{0xde,0x80},
};

static const struct sensor_reg sp0a8_contrast_plus_1_regs[] = 
{
	{0xfd,0x00},
	{0xdd,0x70},
	{0xde,0x90},
};

static const struct sensor_reg sp0a8_contrast_plus_2_regs[] =
{
	{0xfd,0x00},
	{0xdd,0x70},
	{0xde,0xa0},
};
/* ========================================================== */
/*  contrast(5) - END */
/* ========================================================== */



/* ========================================================== */
/*  sharpness(5) -START */
/* ========================================================== */

static const struct sensor_reg sp0a8_sharpness_minus_3_regs[] = {
	{0xfd,0x01},
	{0xe9,0x10},
	{0xed,0x10},
};

static const struct sensor_reg sp0a8_sharpness_minus_2_regs[] = {
	{0xfd,0x01},
	{0xe9,0x10},
	{0xed,0x20},
};

static const struct sensor_reg sp0a8_sharpness_minus_1_regs[] = {
	{0xfd,0x01},
	{0xe9,0x20},
	{0xed,0x20},
};

static const struct sensor_reg sp0a8_sharpness_default_regs[] = {
	{0xfd,0x01},
	{0xe9,0x20},
	{0xed,0x30},
};

static const struct sensor_reg sp0a8_sharpness_plus_1_regs[] = {
	{0xfd,0x01},
	{0xe9,0x30},
	{0xed,0x30},
};

static const struct sensor_reg sp0a8_sharpness_plus_2_regs[] = {
	{0xfd,0x01},
	{0xe9,0x30},
	{0xed,0x40},
};

static const struct sensor_reg sp0a8_sharpness_plus_3_regs[] = {
	{0xfd,0x01},
	{0xe9,0x40},
	{0xed,0x40},
};
/* ========================================================== */
/*  sharpness(5) -END */
/* ========================================================== */


/* ========================================================== */
/*  saturation(5) - START */
/* ========================================================== */

static const struct sensor_reg sp0a8_saturation_minus_2_regs[] = {
	{0xfd,0x00},
	{0xd7,0x70},
	{0xd8,0x70},
	{0xd9,0x70},
	{0xda,0x40},
};

static const struct sensor_reg sp0a8_saturation_minus_1_regs[] = {
	{0xfd,0x00},
	{0xd7,0x70},
	{0xd8,0x70},
	{0xd9,0x70},
	{0xda,0x50},
};

static const struct sensor_reg sp0a8_saturation_default_regs[] = {
	{0xfd,0x00},
	{0xd7,0x70},
	{0xd8,0x70},
	{0xd9,0x70},
	{0xda,0x60},
};

static const struct sensor_reg sp0a8_saturation_plus_1_regs[] = {
	{0xfd,0x00},
	{0xd7,0x70},
	{0xd8,0x70},
	{0xd9,0x70},
	{0xda,0x70},
};

static const struct sensor_reg sp0a8_saturation_plus_2_regs[] = {
	{0xfd,0x00},
	{0xd7,0x70},
	{0xd8,0x70},
	{0xd9,0x70},
	{0xda,0x80},
};
/* ========================================================== */
/*  saturation(5) - END */
/* ========================================================== */

/* ========================================================== */
/* FPS() */
/* ========================================================== */

static const struct sensor_reg sp0a8_fps_auto_regs[] = {

};


static const struct sensor_reg sp0a8_fps_5_regs[] = {
};
static const struct sensor_reg sp0a8_fps_7_regs[] = {

};

static const struct sensor_reg sp0a8_fps_10_regs[] = {
};

static const struct sensor_reg sp0a8_fps_15_regs[] = {

};

static const struct sensor_reg sp0a8_fps_20_regs[] = {

};
static const struct sensor_reg sp0a8_fps_25_regs[] = {
};

static const struct sensor_reg sp0a8_fps_30_regs[] = {

};


/* ========================================================== */
/* FPS()  - END */
/* ========================================================== */



/* ========================================================== */
/* preivew size() - START */
/* ========================================================== */
static const struct sensor_reg sp0a8_preview_size_176x144_regs[] = {

};

static const struct sensor_reg sp0a8_preview_size_320x240_regs[] = {

};

static const struct sensor_reg sp0a8_preview_size_352x288_regs[] = {

};

static const struct sensor_reg sp0a8_preview_size_640x480_regs[] = {

};


/* ========================================================== */
/* PREVIEW SIZE - END */
/* ========================================================== */



/* ========================================================== */
/* CAPTURE START */
/* ========================================================== */

static const struct sensor_reg sp0a8_capture_size_176x144_regs[] = {

};

static const struct sensor_reg sp0a8_capture_size_320x240_regs[] = {

};

static const struct sensor_reg sp0a8_capture_size_352x288_regs[] = {

};

static const struct sensor_reg sp0a8_capture_size_640x480_regs[] = {

};



/* ========================================================== */
/* CAPTURE SIZE - END */
/* ========================================================== */





/* ========================================================== */
/* AWB - START */
/* ========================================================== */

#endif /* __CAMDRV_SP0A28_H__ */
