/*===========================================================================

	Copyright (c) 2011, QUALCOMM Incorporated. All Rights Reserved.
	QUALCOMM Proprietary and Confidential.

============================================================================*/
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include "sensor.h"
#include "config.h"
#include "sp0a28_u.h"
#include "camera_dbg.h"

/* CAMIF output resolutions */
#define SP0A28_FULL_WIDTH                 640
#define SP0A28_FULL_HEIGHT                480
#define SP0A28_QTR_WIDTH                  640
#define SP0A28_QTR_HEIGHT                 480

/* 816x612, 24MHz MCLK 96MHz PCLK */
uint32_t SP0A28_FULL_SIZE_DUMMY_PIXELS   = 0;
uint32_t SP0A28_FULL_SIZE_DUMMY_LINES    = 0;
uint32_t SP0A28_FULL_SIZE_WIDTH          = 640;
uint32_t SP0A28_FULL_SIZE_HEIGHT         = 480;

uint32_t SP0A28_QTR_SIZE_DUMMY_PIXELS    = 0;
uint32_t SP0A28_QTR_SIZE_DUMMY_LINES     = 0;
uint32_t SP0A28_QTR_SIZE_WIDTH           = 640;
uint32_t SP0A28_QTR_SIZE_HEIGHT          = 480;

uint32_t SP0A28_HRZ_FULL_BLK_PIXELS      = 16;
uint32_t SP0A28_VER_FULL_BLK_LINES       = 12;
uint32_t SP0A28_HRZ_QTR_BLK_PIXELS       = 16;
uint32_t SP0A28_VER_QTR_BLK_LINES        = 12;

#define SP0A28_FOCAL_LENGTH               1.15f
#define SP0A28_HORIZONTAL_VIEW_ANGLE      54.8f
#define SP0A28_VERTICAL_VIEW_ANGLE        42.5f


/* Sensor model number used in camear for this sensor */
#define SENSOR_MODEL_NO_SP0A28 "sp0a28"
#define SP0A28_LOAD_CHROMATIX(n)	"libchromatix_"SENSOR_MODEL_NO_SP0A28"_"#n".so"

/* Types of shared object available for this sensor to load dynamically */
char *sp0a28_load_chromatix[SENSOR_LOAD_CHROMATIX_MAX] = {
	NULL, /* camera / camcorder preview */
	NULL, /* Non HD Video recording */
	NULL, /* HD video recording */
	NULL, /* HFR video recording */
	NULL, /* AR */
	NULL, /* ZSL */
};

#if 0//debug
  #undef CDBG
  #define CDBG LOGE
#endif

/*****************************************************************************
 *                          RUN TIME VARIABLES
 ****************************************************************************/

static float sp0a28_register_to_real_gain(uint16_t reg_gain)
{
	float real_gain;

	real_gain = (float)((((float)(reg_gain >> 4) + 1.0) * \
	(16.0 + (float)(reg_gain % 16))) / 16.0);
	return real_gain;
} /* sp0a28_register_to_real_gain */

/*===========================================================================
 * FUNCTION    - sp0a28_write_exposure_gain -
 *
 * DESCRIPTION:
 *==========================================================================*/
static int8_t sp0a28_write_exposure_gain(void *sctrl, uint16_t gain, uint32_t line)
{
	return TRUE;
} /* sp0a28_write_exposure_gain */

/*===========================================================================
 * FUNCTION    - sp0a28_video_config -
 *
 * DESCRIPTION:
 *==========================================================================*/
static int8_t sp0a28_video_config(void *sctrl)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_SET_MODE;
	cfg.mode = SENSOR_PREVIEW_MODE;
	switch (ctrl->sensor.prev_res) {
		case SENSOR_QTR_SIZE:
			{
			/* Sensor output data format */
			ctrl->sensor.format = CAMIF_YCBCR_Y_CB_Y_CR;
			/* Set the current dimensions */
			ctrl->sensor.sensor_width = ctrl->sensor.qtr_size_width;
			ctrl->sensor.sensor_height = ctrl->sensor.qtr_size_height;
			/* CAMIF frame */
			ctrl->sensor.camif_frame_config.pixelsPerLine = (SP0A28_QTR_SIZE_WIDTH + SP0A28_QTR_SIZE_DUMMY_PIXELS)*2;
			ctrl->sensor.camif_frame_config.linesPerFrame = SP0A28_QTR_SIZE_HEIGHT + SP0A28_QTR_SIZE_DUMMY_LINES;
			/* CAMIF window */
			ctrl->sensor.camif_window_width_config.firstPixel = SP0A28_QTR_SIZE_DUMMY_PIXELS;
			ctrl->sensor.camif_window_width_config.lastPixel = ctrl->sensor.camif_window_width_config.firstPixel + (ctrl->sensor.sensor_width*2) - 1;
			ctrl->sensor.camif_window_height_config.firstLine = SP0A28_QTR_SIZE_DUMMY_LINES;
			ctrl->sensor.camif_window_height_config.lastLine = ctrl->sensor.camif_window_height_config.firstLine + ctrl->sensor.sensor_height - 1;
			cfg.rs = SENSOR_QTR_SIZE;
			}/* CAMSENSOR_QTR_SIZE */
			break;
		case SENSOR_FULL_SIZE:
			{
			/* Sensor output data format */
			ctrl->sensor.format = CAMIF_YCBCR_Y_CB_Y_CR;
			ctrl->sensor.sensor_width = ctrl->sensor.full_size_width;
			ctrl->sensor.sensor_height = ctrl->sensor.full_size_height;
			/* CAMIF frame */
			ctrl->sensor.camif_frame_config.pixelsPerLine = SP0A28_FULL_SIZE_WIDTH*2;
			ctrl->sensor.camif_frame_config.linesPerFrame = SP0A28_FULL_SIZE_HEIGHT;
			/* CAMIF window */
			ctrl->sensor.camif_window_width_config.firstPixel = SP0A28_FULL_SIZE_DUMMY_PIXELS;
			ctrl->sensor.camif_window_width_config.lastPixel = ctrl->sensor.camif_window_width_config.firstPixel + (ctrl->sensor.sensor_width*2) - 1;
			ctrl->sensor.camif_window_height_config.firstLine = SP0A28_FULL_SIZE_DUMMY_LINES;
			ctrl->sensor.camif_window_height_config.lastLine = ctrl->sensor.camif_window_height_config.firstLine + ctrl->sensor.sensor_height - 1;
			cfg.rs = SENSOR_FULL_SIZE;
			}/* CAMSENSOR_FULL_SIZE */
			break;
		default:
			return FALSE;
	}/* switch */
	if (ioctl(ctrl->sfd, MSM_CAM_IOCTL_SENSOR_IO_CFG, &cfg) < 0) {
		CDBG("sp0a28 failed %d\n", __LINE__);
		return FALSE;
	}
	ctrl->sensor.curr_res = ctrl->sensor.prev_res;
	ctrl->sensor.dynamic_params.snapshot_changed_fps = FALSE;
	/* Disable the Epoch Interrupt by setting more
	* number of lines than in a frame */
	ctrl->sensor.epoch_line = 10000;
	return TRUE;
}/* sp0a28_video_config */

/*===========================================================================
 * FUNCTION    - sp0a28_real_to_register_gain -
 *
 * DESCRIPTION:
 *==========================================================================*/
static uint16_t sp0a28_real_to_register_gain(float gain)
{
	return 0;
}/* sp0a28_real_to_register_gain */

/*===========================================================================
 * FUNCTION    - sp0a28_snapshot_config -
 *
 * DESCRIPTION:
 *==========================================================================*/
static int8_t sp0a28_snapshot_config(void *sctrl)
{
	float my_preview_gain = 0;
	uint32_t snap_line_count;
	uint32_t snap_reg_gain;
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_SET_MODE;
	cfg.mode = SENSOR_SNAPSHOT_MODE;
	ctrl->sensor.format = CAMIF_YCBCR_Y_CB_Y_CR;
	/* Select discard first frame. */
	ctrl->sensor.discardFirstFrame = FALSE;
	/* Set the current dimensions */
	ctrl->sensor.sensor_width = ctrl->sensor.full_size_width;
	ctrl->sensor.sensor_height = ctrl->sensor.full_size_height;
	/* CAMIF frame */
	ctrl->sensor.camif_frame_config.pixelsPerLine = SP0A28_FULL_SIZE_WIDTH * 2;
	ctrl->sensor.camif_frame_config.linesPerFrame = SP0A28_FULL_SIZE_HEIGHT;
	/* CAMIF window */
	ctrl->sensor.camif_window_width_config.firstPixel = SP0A28_FULL_SIZE_DUMMY_PIXELS;
	ctrl->sensor.camif_window_width_config.lastPixel = ctrl->sensor.camif_window_width_config.firstPixel + (ctrl->sensor.sensor_width * 2) - 1;
	ctrl->sensor.camif_window_height_config.firstLine = SP0A28_FULL_SIZE_DUMMY_LINES;
	ctrl->sensor.camif_window_height_config.lastLine = ctrl->sensor.camif_window_height_config.firstLine + ctrl->sensor.sensor_height - 1;

	if (ioctl(ctrl->sfd, MSM_CAM_IOCTL_SENSOR_IO_CFG, &cfg) < 0) {
		CDBG("sp0a28 failed %d\n", __LINE__);
		return FALSE;
	}
	ctrl->sensor.curr_res = ctrl->sensor.pict_res;
#ifdef FEATURE_STROBE_FLASH
	ctrl->sensor.epoch_line = ctrl->sensor->camif_window_height_config.lastLine;
#else
	/* Disable the Epoch Interrupt by setting more number of lines
	than in a frame*/
	ctrl->sensor.epoch_line = 0xFFFF;
#endif
	ctrl->sensor.dynamic_params.snapshot_pixelsPerLine = ctrl->sensor.camif_frame_config.pixelsPerLine + SP0A28_HRZ_FULL_BLK_PIXELS;
	ctrl->sensor.dynamic_params.snapshot_linesPerFrame = ctrl->sensor.camif_frame_config.linesPerFrame + SP0A28_VER_FULL_BLK_LINES;
	return TRUE;
}/* sp0a28_snapshot_config */

/*===========================================================================
 * FUNCTION    - sp0a28_raw_snapshot_config -
 *
 * DESCRIPTION:
 *==========================================================================*/
static int8_t sp0a28_raw_snapshot_config(void *sctrl)
{
	float my_preview_gain = 0;
	uint32_t snap_line_count;
	uint32_t snap_reg_gain;
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_SET_MODE;
	cfg.mode = SENSOR_SNAPSHOT_MODE;
	ctrl->sensor.format = CAMIF_YCBCR_Y_CB_Y_CR;
	/* Select discard first frame. */
	ctrl->sensor.discardFirstFrame = TRUE;
	/* Set the current dimensions */
	ctrl->sensor.sensor_width = ctrl->sensor.full_size_width;
	ctrl->sensor.sensor_height = ctrl->sensor.full_size_height;
	/* CAMIF frame */
	ctrl->sensor.camif_frame_config.pixelsPerLine = SP0A28_FULL_SIZE_WIDTH * 2;
	ctrl->sensor.camif_frame_config.linesPerFrame = SP0A28_FULL_SIZE_HEIGHT;
	/* CAMIF window */
	ctrl->sensor.camif_window_width_config.firstPixel = SP0A28_FULL_SIZE_DUMMY_PIXELS;
	ctrl->sensor.camif_window_width_config.lastPixel = ctrl->sensor.camif_window_width_config.firstPixel + (ctrl->sensor.sensor_width * 2) - 1;
	ctrl->sensor.camif_window_height_config.firstLine = SP0A28_FULL_SIZE_DUMMY_LINES;
	ctrl->sensor.camif_window_height_config.lastLine = ctrl->sensor.camif_window_height_config.firstLine + ctrl->sensor.sensor_height - 1;
	if (ioctl(ctrl->sfd, MSM_CAM_IOCTL_SENSOR_IO_CFG, &cfg) < 0) {
		CDBG("sp0a28 failed %d\n", __LINE__);
		return FALSE;
	}
	ctrl->sensor.curr_res = ctrl->sensor.pict_res;
	ctrl->sensor.camif_epoch_intr.epoch_line = 1192;
	return TRUE;
} /* sp0a28_raw_snapshot_config */

/*===========================================================================
 * FUNCTION    - sp0a28_power_down -
 *
 * DESCRIPTION:
 *==========================================================================*/
static int8_t sp0a28_power_down(void *sctrl)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_PWR_DOWN;
	return TRUE;
} /* sp0a28_power_down */

/*===========================================================================
 * FUNCTION    - sp0a28_set_default_focus -
 *
 * DESCRIPTION:
 *==========================================================================*/
static int8_t sp0a28_set_default_focus(void *sctrl, int32_t af_step)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_SET_DEFAULT_FOCUS;
	cfg.cfg.focus.steps = af_step;
	return TRUE;
}/* sp0a28_set_default_focus */

/*===========================================================================
 * FUNCTION    - sp0a28_move_focus -
 *
 * DESCRIPTION:
 *==========================================================================*/
static int8_t sp0a28_move_focus(void *sctrl, int direction, int32_t num_steps)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_MOVE_FOCUS;
	cfg.cfg.focus.dir = direction;
	cfg.cfg.focus.steps = num_steps;
	return TRUE;
} /* sp0a28_move_focus */

/*===========================================================================
 * FUNCTION    - sp0a28_set_frame_rate -
 *
 * DESCRIPTION:
 *==========================================================================*/
static int8_t sp0a28_set_frame_rate(void *sctrl, uint16_t fps)
{
	/*Needs to enable for camcorder */
	uint32_t frame_mult = 1 * Q10;
	uint32_t fps_divider = 1 * Q10;
	uint32_t snapshot_fps_divider = 1 * Q10;
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	CDBG("%s\n", __func__);
	if (ctrl->sfd <= 0)
		return FALSE;
	fps_divider = (ctrl->sensor.max_preview_fps * Q10) / fps;
	if (fps_divider == (1 * Q10))
		snapshot_fps_divider = 1 * Q10;
	else
		snapshot_fps_divider = 1 * Q10;
	/*Preview mode update only */
	if (ctrl->sensor.curr_res == ctrl->sensor.prev_res)
		frame_mult = fps_divider;
	else
		frame_mult = snapshot_fps_divider;
	/*use normal exposure time for antibanding to work properly. */
	if (fps == 30 * Q8)
		fps_divider = 1 * Q10;
	cfg.cfgtype = CFG_SET_FPS;
	/* This fps is needed in Kernel space for the sp0a28 sensor*/
	cfg.cfg.fps.f_mult = frame_mult;
	cfg.cfg.fps.fps_div = fps_divider;
	cfg.cfg.fps.pict_fps_div = snapshot_fps_divider;
	if (ctrl->sensor.curr_res == ctrl->sensor.prev_res)
		ctrl->sensor.current_fps = fps;
	return TRUE;
}/* sp0a28_set_frame_rate */

/*===========================================================================
 * FUNCTION    - sp0a28_get_preview_lines_per_frame -
 *
 * DESCRIPTION:
 *==========================================================================*/
static uint16_t sp0a28_get_preview_lines_per_frame(void *sctrl)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_GET_PREV_L_PF;
	cfg.cfg.prevl_pf = 0;
	return(cfg.cfg.prevl_pf);
}

/*===========================================================================
 * FUNCTION    - x -
 *
 * DESCRIPTION:
 *==========================================================================*/
static uint16_t sp0a28_get_preview_pixels_per_line(void *sctrl)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_GET_PREV_P_PL;
	cfg.cfg.prevp_pl = 0;
	return(cfg.cfg.prevp_pl);
} /* sp0a28_get_preview_pixels_per_line */

/*===========================================================================
 * FUNCTION    - sp0a28_get_snapshot_lines_per_frame -
 *
 * DESCRIPTION:
 *==========================================================================*/
static uint16_t sp0a28_get_snapshot_lines_per_frame(void *sctrl)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_GET_PICT_L_PF;
	cfg.cfg.pictl_pf = 0;
	return(cfg.cfg.pictl_pf);
} /*sp0a28_get_snapshot_lines_per_frame*/

/*===========================================================================
 * FUNCTION    - sp0a28_get_snapshot_pixels_per_line -
 *
 * DESCRIPTION:
 *==========================================================================*/
static uint16_t sp0a28_get_snapshot_pixels_per_line(void *sctrl)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_GET_PICT_P_PL;
	cfg.cfg.pictp_pl = 0;
	return(cfg.cfg.pictp_pl);
} /*sp0a28_get_snapshot_pixels_per_line*/

/*===========================================================================
 * FUNCTION    - sp0a28_get_snapshot_fps -
 *
 * DESCRIPTION:
 *==========================================================================*/
static uint16_t sp0a28_get_snapshot_fps(void *sctrl, uint16_t fps)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_GET_PICT_FPS;
	cfg.cfg.gfps.prevfps = fps;
	cfg.cfg.gfps.pictfps = 0;
	return(cfg.cfg.gfps.pictfps);
} /*sp0a28_get_snapshot_fps*/

/*===========================================================================
 * FUNCTION    - sp0a28_get_snapshot_max_exposure_line_count -
 *
 * DESCRIPTION:
 *==========================================================================*/
static uint32_t sp0a28_get_snapshot_max_exposure_line_count(void *sctrl)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_GET_PICT_MAX_EXP_LC;
	cfg.cfg.pict_max_exp_lc = 0;
	return(cfg.cfg.pict_max_exp_lc);
} /*sp0a28_get_snapshot_max_exposure_line_count*/

/*===========================================================================
 * FUNCTION    - sp0a28_set_snapshot_frame_rate -
 *
 * DESCRIPTION:
 *==========================================================================*/
static int8_t sp0a28_set_snapshot_frame_rate(void *sctrl, uint16_t fps)
{
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	if (sp0a28_set_frame_rate(sctrl, fps) == FALSE)
		return FALSE;
	ctrl->sensor.dynamic_params.snapshot_changed_fps = TRUE;
	return TRUE;
} /*sp0a28_set_snapshot_frame_rate*/

/*===========================================================================
 * FUNCTION    - sp0a28_set_lens_shading_correction -
 *
 * DESCRIPTION:
 *==========================================================================*/
int8_t sp0a28_set_lens_shading_correction(void *sctrl, int8_t lens_shading_status)
{
	printf("!!!Sensor Lens Shading is not supported on 3MP BAYER sensor!!!\n");
	usleep(1000 * 50);
	return TRUE;
} /* sp0a28_set_lens_shading_correction */

/*===========================================================================
 * FUNCTION    - sp0a28_set_snapshot_exposure_gain -
 *
 * DESCRIPTION:
 *==========================================================================*/
static int8_t sp0a28_set_snapshot_exposure_gain(void *sctrl, void *snapshot_config)
{
	float snap_gain = 0;
	int8_t rc = 0;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;

	uint32_t snapshot_wait = 2 * Q8 * 1000 / (sp0a28_get_snapshot_fps(ctrl, ctrl->sensor.current_fps));
	sensor_snapshot_config_t *snapshot_config_ptr = (sensor_snapshot_config_t *) snapshot_config;
	rc = sp0a28_write_exposure_gain(sctrl,snapshot_config_ptr->gain, snapshot_config_ptr->linecount);
	usleep(snapshot_wait* 1000);
	return (rc);
} /* sp0a28_set_snapshot_exposure_gain */

/*=========================================================================
 * FUNCTION    - sp0a28_setup_camctrl_tbl -
 *
 * DESCRIPTION:
 *==========================================================================*/
static void sp0a28_setup_camctrl_tbl(sensor_camctrl_tbl_t *camctrl_tbl_ptr)
{
	/* Exposure Related 3A function */
	camctrl_tbl_ptr->fast_convergence_is_supported = FALSE;

	/* 3A handler for nightshot mode */
	camctrl_tbl_ptr->nightshot_is_supported = FALSE;

	/* Set camctrl_tbl static variables */
	camctrl_tbl_ptr->iso100_gain = (uint32_t) (1.87 * Q8);

	/* AF common tuning parameters -- start */
	camctrl_tbl_ptr->af_is_supported = FALSE;
	camctrl_tbl_ptr->hjr_bayer_filtering_enable = FALSE;

	/* AEC common tuning parameter -- start */
	camctrl_tbl_ptr->high_luma_region_threshold = 130;
	/* AEC common tuning parameter -- end */

	/* AWB parameters */
	camctrl_tbl_ptr->noon_rg = 0.674;
	camctrl_tbl_ptr->noon_bg = 0.862;
} /*sp0a28_setup_camctrl_tbl*/

/*===========================================================================
 * FUNCTION    - sp0a28_set_exposure_compensation -
 *
 * DESCRIPTION:
 *==========================================================================*/
static uint8_t sp0a28_set_exposure_compensation(void *sctrl, int8_t compensation)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *)sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_SET_EXPOSURE_COMPENSATION;
	cfg.cfg.exp_compensation = compensation;
	if (ioctl(ctrl->sfd, MSM_CAM_IOCTL_SENSOR_IO_CFG, &cfg) < 0)
	{
		LOGE("sp0a28 failed %d\n", __LINE__);
		return FALSE;
	}
	return TRUE;
}

/*===========================================================================
 * FUNCTION    - sp0a28_set_antibanding -
 *
 * DESCRIPTION:
 *==========================================================================*/
static int8_t sp0a28_set_antibanding(void *sctrl, int8_t antibanding)
{
	camera_antibanding_type temp;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;
	struct sensor_cfg_data cfg;

	temp = (camera_antibanding_type) antibanding;
	LOGE("--CAMERA-- %s in!\n",__func__);
	if (ctrl->sfd <= 0)
		return FALSE;
	LOGE("Entered sp0a28_set_antibanding, ctrl->sfd = %d, op_mode = %d\n", ctrl->sfd, ctrl->sensor.op_mode);
	cfg.cfgtype = CFG_SET_ANTIBANDING;
	cfg.cfg.antibanding = antibanding;
	LOGE("Setting antibanding to %d\n", antibanding);
	if (ioctl(ctrl->sfd, MSM_CAM_IOCTL_SENSOR_IO_CFG, &cfg))
		return FALSE;
	ctrl->sensor.antibanding_enable = antibanding;
	return TRUE;
}

/*===========================================================================
 * FUNCTION    - sp0a28_set_saturation -
 *
 * DESCRIPTION:
 *==========================================================================*/
static uint8_t sp0a28_set_saturation(void *sctrl, int8_t saturation)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *)sctrl;

	LOGE("--CAMERA-- %s in!\n",__func__);
	LOGE("Entered sp0a28_set_saturation, ctrl->sfd = %d\n", ctrl->sfd);
	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_SET_SATURATION;
	cfg.cfg.saturation = saturation;
	if (ioctl(ctrl->sfd, MSM_CAM_IOCTL_SENSOR_IO_CFG, &cfg) < 0) {
		LOGE("sp0a28 failed %d\n", __LINE__);
		return FALSE;
	}
	return TRUE;
}

/*===========================================================================
 * FUNCTION    - sp0a28_get_effects_supported -
 *
 * DESCRIPTION:
 *==========================================================================*/

static int8_t sp0a28_get_effects_supported(uint32_t* effects)
{
	if (NULL == effects)
		return FALSE;

	*effects = (1<<CAMERA_EFFECT_OFF) |
	(1<<CAMERA_EFFECT_MONO) |
	(1<<CAMERA_EFFECT_NEGATIVE) |
	(1<<CAMERA_EFFECT_SOLARIZE) |
	(1<<CAMERA_EFFECT_SEPIA)	|
	(1<<CAMERA_EFFECT_BW)		|
	(1<<CAMERA_EFFECT_BLUISH)	|
	(1<<CAMERA_EFFECT_GREENISH)	|
	(1<<CAMERA_EFFECT_REDDISH);
	return TRUE;
}

/*===========================================================================
 * FUNCTION    - sp0a28_set_effect -
 *
 * DESCRIPTION:
 *==========================================================================*/

static int8_t sp0a28_set_effect(void *sctrl, int8_t effect)
{
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *) sctrl;
	struct sensor_cfg_data cfg;

	CDBG("Entered sp0a28_set_effect, ctrl->sfd = %d, op_mode = %d\n", ctrl->sfd, ctrl->sensor.op_mode);
	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_SET_EFFECT;
	switch (ctrl->sensor.op_mode) {
		case SENSOR_OP_MODE_PREVIEW:
			cfg.mode = SENSOR_PREVIEW_MODE;
			break;
		case SENSOR_OP_MODE_SNAPSHOT:
			cfg.mode = SENSOR_SNAPSHOT_MODE;
			break;
		case SENSOR_OP_MODE_HFR_60FPS:
			cfg.mode = SENSOR_HFR_60FPS_MODE;
			break;
		case SENSOR_OP_MODE_HFR_90FPS:
			cfg.mode = SENSOR_HFR_90FPS_MODE;
			break;
		case SENSOR_OP_MODE_HFR_120FPS:
			cfg.mode = SENSOR_HFR_120FPS_MODE;
			break;
		case SENSOR_RAW_SNAPSHOT_MODE:
			return TRUE;
		default:
			return FALSE;
	}

	cfg.cfg.effect = effect;
	CDBG("Setting effect to %d\n", effect);
	if (ioctl(ctrl->sfd, MSM_CAM_IOCTL_SENSOR_IO_CFG, &cfg))
		return FALSE;
	ctrl->sensor.seffect = effect;
	return TRUE;
}

/*===========================================================================
 * FUNCTION    - sp0a28_set_contrast -
 *
 * DESCRIPTION:
 *==========================================================================*/

static uint8_t sp0a28_set_contrast (void *sctrl, int8_t contrast)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *)sctrl;

	LOGE("--CAMERA-- %s in!\n",__func__);
	LOGE("Entered sensor_set_contrast, ctrl->sfd = %d\n", ctrl->sfd);
	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_SET_CONTRAST;
	cfg.cfg.contrast = contrast;
	if (ioctl(ctrl->sfd, MSM_CAM_IOCTL_SENSOR_IO_CFG, &cfg) < 0) {
		LOGE("sp0a28 failed %d\n", __LINE__);
		return FALSE;
	}
	return TRUE;
}

/*===========================================================================
 * FUNCTION    - sp0a28_set_sharpness -
 *
 * DESCRIPTION:
 *==========================================================================*/
static uint8_t sp0a28_set_sharpness(void *sctrl, int8_t sharpness)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *)sctrl;

	LOGE("--CAMERA-- %s in!\n",__func__);
	LOGE("Entered sp0a28_set_sharpness, ctrl->sfd = %d\n", ctrl->sfd);
	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_SET_SHARPNESS;
	cfg.cfg.sharpness = sharpness;
	if (ioctl(ctrl->sfd, MSM_CAM_IOCTL_SENSOR_IO_CFG, &cfg) < 0) {
		LOGE("sp0a28 failed %d\n", __LINE__);
		return FALSE;
	}
	return TRUE;
}

/*===========================================================================
 * FUNCTION    - sp0a28_set_iso -
 *
 * DESCRIPTION:
 *==========================================================================*/
static uint8_t sp0a28_set_iso(void *sctrl, int8_t iso_type)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *)sctrl;

	if (ctrl->sfd <= 0)
		return FALSE;
	cfg.cfgtype = CFG_SET_ISO;
	cfg.cfg.iso_type = iso_type;
	if (ioctl(ctrl->sfd, MSM_CAM_IOCTL_SENSOR_IO_CFG, &cfg) < 0) {
		CDBG("sp0a28 failed %d\n", __LINE__);
		return FALSE;
	}
	return TRUE;
}

static uint8_t sp0a28_set_wb_oem(void *sctrl, int8_t param)
{
	struct sensor_cfg_data cfg;
	sensor_ctrl_t *ctrl = (sensor_ctrl_t *)sctrl;

	CDBG("%s...set %d\n",__func__,param);
	CDBG("--CAMERA-- %s in!\n",__func__);

	if (ctrl->sfd <= 0)
		return FALSE;

	cfg.cfgtype = CFG_SET_WB;
	cfg.cfg.wb_val = param;

	CDBG("set ioctl to camera driver \r\n");
	if (ioctl(ctrl->sfd, MSM_CAM_IOCTL_SENSOR_IO_CFG, &cfg) < 0) {
		CDBG("sp0a28 failed %d\n", __LINE__);
		return FALSE;
	}
	return TRUE;
}

/*===========================================================================
 * FUNCTION    - sp0a28_register -
 *
 * DESCRIPTION:
 *==========================================================================*/
static void sp0a28_register(sensor_function_table_t * table)
{
	table->sensor_video_config = sp0a28_video_config;
	table->sensor_snapshot_config = sp0a28_snapshot_config;
	table->sensor_raw_snapshot_config = sp0a28_raw_snapshot_config;
	table->sensor_power_down = sp0a28_power_down;
	table->sensor_write_exposure_gain = sp0a28_write_exposure_gain;
	table->sensor_set_default_focus = sp0a28_set_default_focus;
	table->sensor_move_focus = sp0a28_move_focus;
	table->sensor_register_to_real_gain = sp0a28_register_to_real_gain;
	table->sensor_real_to_register_gain = sp0a28_real_to_register_gain;
	table->sensor_set_frame_rate = sp0a28_set_frame_rate;
	table->sensor_set_lens_shading_correction = sp0a28_set_lens_shading_correction;
	table->sensor_set_exposure_compensation = sp0a28_set_exposure_compensation;
	table->sensor_set_antibanding = sp0a28_set_antibanding;
	/* Set camctrl_tbl function pointers */
	table->get_preview_lines_per_frame = sp0a28_get_preview_lines_per_frame;
	table->get_preview_pixels_per_line = sp0a28_get_preview_pixels_per_line;
	table->get_snapshot_lines_per_frame = sp0a28_get_snapshot_lines_per_frame;
	table->get_snapshot_pixels_per_line = sp0a28_get_snapshot_pixels_per_line;
	table->get_snapshot_frame_per_sec = sp0a28_get_snapshot_fps;
	table->get_snapshot_max_exposure_line_count = sp0a28_get_snapshot_max_exposure_line_count;
	table->set_snapshot_frame_rate = sp0a28_set_snapshot_frame_rate;
	table->set_snapshot_exposure_gain = sp0a28_set_snapshot_exposure_gain;
	table->sensor_set_saturation = sp0a28_set_saturation;
	table->sensor_get_effects_supported = sp0a28_get_effects_supported;
	table->sensor_set_effect = sp0a28_set_effect;
	table->sensor_set_contrast = sp0a28_set_contrast;
	table->sensor_set_sharpness = sp0a28_set_sharpness;
	table->sensor_set_iso= sp0a28_set_iso;
	table->sensor_get_hfr_supported = NULL;
	table->sensor_set_wb_oem = sp0a28_set_wb_oem;
}/* sp0a28_register */

/*===========================================================================
 * FUNCTION    - sp0a28_process_start -
 *
 * DESCRIPTION:
 *==========================================================================*/
int8_t sp0a28_process_start(void *ctrl)
{
	CDBG("Inside sp0a28_process_start \n");

	sensor_ctrl_t *sensor = (sensor_ctrl_t *) ctrl;
	sensor->sensor.prev_res = SENSOR_FULL_SIZE;
	sensor->sensor.pict_res = SENSOR_FULL_SIZE;

	sensor->sensor.curr_res = sensor->sensor.prev_res;

	/* ------------------  Sensor-specific Config -------------- */
	/* Make/model of sensor */
	sensor->sensor.sensor_model = SENSOR_SP0A28;
	/* CCD or CMOS */
	sensor->sensor.sensor_type = SENSOR_CMOS;
	/* BAYER or YCbCr */
	sensor->sensor.output_format = SENSOR_YCBCR;

	/* Initialize CAMIF operation mode */
	sensor->sensor.camif_config.syncmode = CAMIF_APS;
	sensor->sensor.camif_config.hsyncedge = CAMIF_HIGH;
	sensor->sensor.camif_config.vsyncedge = CAMIF_HIGH;
	sensor->sensor.camif_efs_config.efs_eol = 0x0000;
	sensor->sensor.camif_efs_config.efs_sol = 0x0000;
	sensor->sensor.camif_efs_config.efs_eof = 0x0000;
	sensor->sensor.camif_efs_config.efs_sof = 0x0000;


	sensor->sensor.focal_length = SP0A28_FOCAL_LENGTH;
	sensor->sensor.horizontal_view_angle = SP0A28_HORIZONTAL_VIEW_ANGLE;
	sensor->sensor.vertical_view_angle = SP0A28_VERTICAL_VIEW_ANGLE;


	if (sensor->sensor.prev_res == SENSOR_QTR_SIZE)
	{
		sensor->sensor.max_video_fps = 30 * Q8;
		/* Application assigned FPS in video mode */
		sensor->sensor.video_fps = 30 * Q8;
		/* Snapshot mode operation */
		sensor->sensor.max_preview_fps = 30 * Q8;
		sensor->sensor.nightshot_fps = 5 * Q8;
		sensor->sensor.nightshot_fps = sensor->sensor.nightshot_fps;
		/* May be assigned with max_preview_fps or nightshot_fps. */
		sensor->sensor.preview_fps = 30 * Q8;
		sensor->sensor.current_fps = sensor->sensor.preview_fps;
		/* VFE's perception of Sensor output capability */
		sensor->sensor.full_size_width = SP0A28_FULL_SIZE_WIDTH;
		sensor->sensor.full_size_height = SP0A28_FULL_SIZE_HEIGHT;
		sensor->sensor.qtr_size_width = SP0A28_QTR_SIZE_WIDTH;
		sensor->sensor.qtr_size_height = SP0A28_QTR_SIZE_HEIGHT;
	} else {
		/*full resolution preview support. */
		sensor->sensor.max_video_fps = 30 * Q8;
		/* Application assigned FPS in video mode */
		sensor->sensor.video_fps = 30 * Q8;
		/* Snapshot mode operation */
		sensor->sensor.max_preview_fps = 30 * Q8;
		sensor->sensor.nightshot_fps = 5 * Q8;
		sensor->sensor.nightshot_fps = sensor->sensor.nightshot_fps;
		/* May be assigned with max_preview_fps or nightshot_fps. */
		sensor->sensor.preview_fps = 30 * Q8;
		sensor->sensor.current_fps = sensor->sensor.preview_fps;
		/* VFE's perception of Sensor output capability */
		sensor->sensor.full_size_width = SP0A28_FULL_SIZE_WIDTH;
		sensor->sensor.full_size_height = SP0A28_FULL_SIZE_HEIGHT;
		/* Quarter size dimensions - 1024x768 */
		sensor->sensor.qtr_size_width = SP0A28_FULL_SIZE_WIDTH;
		sensor->sensor.qtr_size_height = SP0A28_FULL_SIZE_HEIGHT;
	}
	switch (sensor->sensor.prev_res) {
		case SENSOR_QTR_SIZE:
			sensor->sensor.preview_dx_decimation =
			sensor->sensor.full_size_width * Q12 / sensor->sensor.qtr_size_width;
			sensor->sensor.preview_dy_decimation =
			sensor->sensor.full_size_height * Q12 / sensor->sensor.qtr_size_height;
			/* Set the current dimensions */
			sensor->sensor.sensor_width = sensor->sensor.qtr_size_width;
			sensor->sensor.sensor_height = sensor->sensor.qtr_size_height;
			break;
		case SENSOR_FULL_SIZE:
			sensor->sensor.preview_dx_decimation = Q12;
			sensor->sensor.preview_dy_decimation = Q12;
			/* Set the current dimensions */
			sensor->sensor.sensor_width = sensor->sensor.full_size_width;
			sensor->sensor.sensor_height = sensor->sensor.full_size_height;
			break;
		default:
			sensor->sensor.preview_dx_decimation = Q12;
			sensor->sensor.preview_dy_decimation = Q12;
			break;
	}
	/* This tells camera service the minimum decimation that is supported
	 * by the sensor.
	 * Ex: if preview is in quarter size mode, then there is a
	 * sensor decimation of 2, so the minimum is 2 */
	sensor->sensor.preview_dx_decimation =
	(sensor->sensor.preview_dx_decimation <
	Q12) ? Q12 : sensor->sensor.preview_dx_decimation;

	sensor->sensor.preview_dy_decimation =
	(sensor->sensor.preview_dy_decimation <
	Q12) ? Q12 : sensor->sensor.preview_dy_decimation;

	/* ------------  Auto Exposure Control Config -------------- */
	sensor->chromatixAnd3a.aec_enable = FALSE;
	sensor->sensor.connection_mode = SENSOR_MIPI_CSI_1;
	sensor->chromatixAnd3a.awb_enable = FALSE;
	sensor->sensor.raw_output_option = SENSOR_8_BIT_DIRECT;

	/* ------------  Auto Frame Rate Config (AFR) -------------- */
	/* A subset of Auto Exposure                                 */
	/* Number of entries or possible frame rates on this sensor
	* Represents the depth used in the array below */

	sensor->sensor.max_video_fps = 30 * Q8;

	/* Application assigned FPS in video mode */
	sensor->sensor.video_fps = 30 * Q8;

	/* Snapshot mode operation */
	sensor->sensor.max_preview_fps = 30 * Q8;
	sensor->sensor.nightshot_fps = 5 * Q8;

	/* May be assigned with max_preview_fps or nightshot_fps. */
	sensor->sensor.preview_fps = 30 * Q8;

	/* -----------------  Anti Banding Config ------------------ */
	sensor->sensor.pixel_clock = 76800000;
	if (sensor->sensor.prev_res == SENSOR_FULL_SIZE) {
		/*pixel_clock_per_line includes all blanking time*/
		sensor->sensor.pixel_clock_per_line = SP0A28_FULL_WIDTH + SP0A28_HRZ_FULL_BLK_PIXELS;
	} else {
		/*pixel_clock_per_line includes all blanking time*/
		sensor->sensor.pixel_clock_per_line = SP0A28_QTR_WIDTH + SP0A28_HRZ_QTR_BLK_PIXELS;
	}
	/* Does the sensor need/use a flash  */
	sensor->sensor.support_auto_flash = FALSE;
	sensor->sensor.pclk_invert = FALSE;
	sensor->sensor.ignore_camif_error = 1;
	sensor->chromatixAnd3a.af_enable = FALSE;
	sensor->chromatixAnd3a.asf_data_3.asf_enable = FALSE;
	sensor->chromatixAnd3a.asf_data_5_preview.asf_enable = FALSE;
	sensor->chromatixAnd3a.asf_data_5_snapshot.asf_enable = FALSE;
	sensor->sensor.seffect = CAMERA_EFFECT_OFF;
	/* Register function table: */
	sp0a28_register(&(sensor->fn_table));
	/* Setup camctrl_tbl */
	sp0a28_setup_camctrl_tbl(&(sensor->camctrl_tbl));
	sensor_post_init(sensor);
	return TRUE;
} /* sp0a28_start */
