/* Copyright (c) 2010-2011, Code Aurora Forum. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#define DEBUG

#include <linux/delay.h>
#include <linux/types.h>
#include <linux/i2c.h>
#include <linux/uaccess.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <media/msm_camera.h>
#include <mach/camera.h>
#include <mach/gpio.h>
#include "sp0a28.h"
#include <mach/vreg.h>

/*=============================================================
    SENSOR REGISTER DEFINES
==============================================================*/
#define Q8    0x00000100

/* Omnivision8810 product ID register address */
#define REG_SP0A28_MODEL_ID                       0X02

#define SP0A28_MODEL_ID                       0XA2
/* Omnivision8810 product ID */

/* Time in milisecs for waiting for the sensor to reset */
#define SP0A28_RESET_DELAY_MSECS    66
#define SP0A28_DEFAULT_CLOCK_RATE   24000000
/* Registers*/

/* Color bar pattern selection */
#define SP0A28_COLOR_BAR_PATTERN_SEL_REG     0x82
/* Color bar enabling control */
#define SP0A28_COLOR_BAR_ENABLE_REG           0x601
/* Time in milisecs for waiting for the sensor to reset*/
#define SP0A28_RESET_DELAY_MSECS    66

extern void front_camera_power_down(void);
extern void front_camera_power_up(void);
extern void front_camera_power_reset(void);

/*============================================================================
                            DATA DECLARATIONS
============================================================================*/


static bool SP0A28_CSI_CONFIG;
/* 816x612, 24MHz MCLK 96MHz PCLK */
uint32_t SP0A28_FULL_SIZE_WIDTH        = 640;
uint32_t SP0A28_FULL_SIZE_HEIGHT       = 480;

uint32_t SP0A28_QTR_SIZE_WIDTH         = 640;
uint32_t SP0A28_QTR_SIZE_HEIGHT        = 480;

uint32_t SP0A28_HRZ_FULL_BLK_PIXELS    = 16;
uint32_t SP0A28_VER_FULL_BLK_LINES     = 12;
uint32_t SP0A28_HRZ_QTR_BLK_PIXELS     = 16;
uint32_t SP0A28_VER_QTR_BLK_LINES      = 12;

struct sp0a28_work_t {
    struct work_struct work;
};
static struct  sp0a28_work_t *sp0a28_sensorw;
static struct  i2c_client *sp0a28_client;
struct sp0a28_ctrl_t {
    const struct  msm_camera_sensor_info *sensordata;
    uint32_t sensormode;
    uint32_t fps_divider;        /* init to 1 * 0x00000400 */
    uint32_t pict_fps_divider;    /* init to 1 * 0x00000400 */
    uint32_t fps;
    int32_t  curr_lens_pos;
    uint32_t curr_step_pos;
    uint32_t my_reg_gain;
    uint32_t my_reg_line_count;
    uint32_t total_lines_per_frame;
    enum sp0a28_resolution_t prev_res;
    enum sp0a28_resolution_t pict_res;
    enum sp0a28_resolution_t curr_res;
    enum sp0a28_test_mode_t  set_test;
    unsigned short imgaddr;
};
static struct sp0a28_ctrl_t *sp0a28_ctrl;
static DECLARE_WAIT_QUEUE_HEAD(sp0a28_wait_queue);
DEFINE_MUTEX(sp0a28_mut);
static int effect_value = 0;
static int16_t sp0a28_effect = CAMERA_EFFECT_OFF;
static unsigned int SAT_U = 0x80;
static unsigned int SAT_V = 0x80;

/*=============================================================*/
#ifdef CONFIG_AW9523
enum {
	AW9523_I2C_GPIO_CAMIF_EN_F = 0, 	/*P0_6: CAMIF_EN_F*/
	AW9523_I2C_GPIO_CAMIF_EN_B, 		/*P0_7: CAMIF_EN_B*/

	AW9523_I2C_GPIO_IO_SWITCH  = 100,	/*P1_0: IO_SWITCH*/
	AW9523_I2C_GPIO_LCD_P_OFF, 			/*P1_1: LCD_P_OFF*/
	AW9523_I2C_GPIO_VREG_LCD_EV_EN, 	/*P1_2: VREG_LCD_EV_EN*/
	AW9523_I2C_GPIO_CHARGE_ON, 			/*P1_3: CHARGE_ON*/
	AW9523_I2C_GPIO_OTG_VDD_OFF,		/*P1_4: OTG_VDD_OFF*/
	AW9523_I2C_GPIO_BT_SYS_RESET_N,		/*P1_5: BT_SYS_RESET_N*/
	AW9523_I2C_GPIO_DTV_IO,				/*P1_6: DTV_IO*/
	AW9523_I2C_GPIO_A_PA_SHD,			/*P1_7: A_PA_SHD*/
};
extern int aw9523_i2c_gpio_output(unsigned char gpio_i2c_type, unsigned char val);
#endif

static int sp0a28_i2c_rxdata(unsigned short saddr, unsigned char *rxdata, int length)
{
    struct i2c_msg msgs[] = {
        {
            .addr  = saddr,
            .flags = 0,
            .len   = 1,
            .buf   = rxdata,
        },
        {
            .addr  = saddr,
            .flags = I2C_M_RD,
            .len   = 1,
            .buf   = rxdata,
        },
    };
    if (i2c_transfer(sp0a28_client->adapter, msgs, 2) < 0) {
        CDBG("sp0a28_i2c_rxdata failed!\n");
        return -EIO;
    }
    return 0;
}
static int32_t sp0a28_i2c_txdata(unsigned short saddr, unsigned char *txdata, int length)
{
    struct i2c_msg msg[] = {
        {
            .addr = saddr,
            .flags = 0,
            .len = 2,
            .buf = txdata,
         },
    };
    if (i2c_transfer(sp0a28_client->adapter, msg, 1) < 0) {
        CDBG("sp0a28_i2c_txdata faild 0x%x\n", sp0a28_client->addr);
        return -EIO;
    }

    return 0;
}

static int32_t sp0a28_i2c_read(uint8_t raddr, uint8_t *rdata, int rlen)
{
    int32_t rc = 0;
    unsigned char buf[1];
    if (!rdata)
        return -EIO;
    memset(buf, 0, sizeof(buf));
    buf[0] = raddr;
    rc = sp0a28_i2c_rxdata(sp0a28_client->addr >> 1, buf, rlen);
    if (rc < 0) {
        CDBG("sp0a28_i2c_read 0x%x failed!\n", raddr);
        return rc;
    }
    *rdata = buf[0];
    return rc;
}
static int32_t sp0a28_i2c_write(uint8_t waddr, uint8_t bdata)
{
    int32_t rc = -EFAULT;
    unsigned char buf[2];
    memset(buf, 0, sizeof(buf));
    buf[0] = waddr;
    buf[1] = bdata;
    //CDBG("i2c_write_b addr = 0x%x, val = 0x%x\n", waddr, bdata);
    rc = sp0a28_i2c_txdata(sp0a28_client->addr >> 1, buf, 2);
    if (rc < 0)
        CDBG("i2c_write_b failed, addr = 0x%x, val = 0x%x!\n",
            waddr, bdata);
    return rc;
}

static int32_t SP0A28_WritePRegs(PSP0A28_WREG pTb, int32_t len)
{
    int32_t i, ret = 0;
    for (i = 0; i < len; i++)
    {
		sp0a28_i2c_write(pTb[i].addr,pTb[i].data);
    }
    return ret;
}

static int32_t sp0a28_sensor_setting(int update_type, int rt)
{
    int32_t i, array_length;
    int32_t rc = 0;
    struct msm_camera_csi_params sp0a28_csi_params;

    CDBG("%s: rt = %d\n", __func__, rt);

    switch (update_type) {
    case REG_INIT:
        SP0A28_CSI_CONFIG = 0;
        //sp0a28_i2c_write(0x0e, 0x08);
        return rc;
        break;
    case UPDATE_PERIODIC:
        if (!SP0A28_CSI_CONFIG) {
            sp0a28_csi_params.lane_cnt = 1;
            sp0a28_csi_params.data_format = CSI_8BIT;
            sp0a28_csi_params.lane_assign = 0xe4;
            sp0a28_csi_params.dpcm_scheme = 0;
            sp0a28_csi_params.settle_cnt = 0x09;
			rc = msm_camio_csi_config(&sp0a28_csi_params);
			msleep(100);
			//??sp0a28_i2c_write(0x0e, 0x00);//open output

            array_length = sizeof(sp0a28_init_settings_array) / sizeof(sp0a28_init_settings_array[0]);
            for (i = 0; i < array_length; i++) {
                rc = sp0a28_i2c_write(sp0a28_init_settings_array[i].reg_addr,sp0a28_init_settings_array[i].reg_val);
                if (rc < 0)
                    return rc;
            }
			msleep(300);
            SP0A28_CSI_CONFIG = 1;
            msleep(20);
            return rc;
        }
#if 0
		if (rt == RES_CAPTURE)
		{
			//??rc = SP0A28Core_WritePREG(gc0339_preview_tbl_30fps);
			if (rc < 0) {
			  CDBG("sensor preview setting failed\n");
			  return rc;
			}
		}
		else//RES_PREVIEW
		{
			//??rc = SP0A28Core_WritePREG(gc0339_capture_tbl);
			if (rc < 0) {
			  CDBG("sensor capture setting failed\n");
			  return rc;
			}
		}
#endif
        break;
    default:
        rc = -EINVAL;
        break;
    }
    return rc;
}

static int32_t sp0a28_video_config(int mode)
{
    int32_t rc = 0;
    int rt;
    /* change sensor resolution if needed */
    rt = RES_PREVIEW;

    CDBG("%s\n", __func__);

    if (sp0a28_sensor_setting(UPDATE_PERIODIC, rt) < 0)
        return rc;
    sp0a28_ctrl->curr_res = sp0a28_ctrl->prev_res;
    sp0a28_ctrl->sensormode = mode;
    return rc;
}

static int32_t sp0a28_snapshot_config(int mode)
{
    int32_t rc = 0;
    int rt;
    /* change sensor resolution if needed */
    rt = RES_CAPTURE;

    CDBG("%s\n", __func__);

    if (sp0a28_sensor_setting(UPDATE_PERIODIC, rt) < 0)
        return rc;
    sp0a28_ctrl->curr_res = sp0a28_ctrl->pict_res;
    sp0a28_ctrl->sensormode = mode;
    return rc;
}


static int32_t sp0a28_set_sensor_mode(int mode,
    int res)
{
    int32_t rc = 0;
    switch (mode) {
    case SENSOR_PREVIEW_MODE:
        rc = sp0a28_video_config(mode);
        break;
    case SENSOR_SNAPSHOT_MODE:
    case SENSOR_RAW_SNAPSHOT_MODE:
        rc = sp0a28_snapshot_config(mode);
        break;
    default:
        rc = -EINVAL;
        break;
    }
    return rc;
}

static int sp0a28_set_exposure_compensation(int compensation)
{
	long rc = 0; 
	
	CDBG("--CAMERA-- %s ...(Start)\n",__func__);
	CDBG("--CAMERA-- %s ...exposure_compensation = %d\n",__func__ , compensation);
	switch(compensation)
	{
		case CAMERA_EXPOSURE_COMPENSATION_LV0:
			CDBG("--CAMERA--CAMERA_EXPOSURE_COMPENSATION_LV0\n");
			rc = SP0A28Core_WritePREG(sp0a28_exposure_compensation_lv0_tbl);
			break;
		case CAMERA_EXPOSURE_COMPENSATION_LV1:
			CDBG("--CAMERA--CAMERA_EXPOSURE_COMPENSATION_LV1\n");
			rc = SP0A28Core_WritePREG(sp0a28_exposure_compensation_lv1_tbl);
			break;
		case CAMERA_EXPOSURE_COMPENSATION_LV2:
			CDBG("--CAMERA--CAMERA_EXPOSURE_COMPENSATION_LV2\n");
			rc = SP0A28Core_WritePREG(sp0a28_exposure_compensation_lv2_default_tbl);
			break;
		case CAMERA_EXPOSURE_COMPENSATION_LV3:
			CDBG("--CAMERA--CAMERA_EXPOSURE_COMPENSATION_LV3\n");
			rc = SP0A28Core_WritePREG(sp0a28_exposure_compensation_lv3_tbl);
			break;
		case CAMERA_EXPOSURE_COMPENSATION_LV4:
			CDBG("--CAMERA--CAMERA_EXPOSURE_COMPENSATION_LV3\n");
			rc = SP0A28Core_WritePREG(sp0a28_exposure_compensation_lv4_tbl);
			break;
		default:
			CDBG("--CAMERA--ERROR CAMERA_EXPOSURE_COMPENSATION\n");
			break;
	}
	CDBG("--CAMERA-- %s ...(End)\n",__func__);
    return rc;
}

static long sp0a28_set_antibanding(int antibanding)
{
	long rc = 0;
	uint8_t temp = 0;
	
	
	sp0a28_i2c_write(0xfd,0x00);
	sp0a28_i2c_read(0x32, &temp, 1);


	CDBG("--CAMERA-- %s ...(Start)\n", __func__);
	CDBG("--CAMERA-- %s ...antibanding = %d\n", __func__, antibanding);
	switch (antibanding)
	{
		case CAMERA_ANTIBANDING_OFF:
			CDBG("--CAMERA--CAMERA_ANTIBANDING_OFF\n");
			break;
		case CAMERA_ANTIBANDING_60HZ:
			CDBG("--CAMERA--CAMERA_ANTIBANDING_60HZ\n");
			rc = SP0A28Core_WritePREG(sp0a28_antibanding_60z_tbl);	
			//sp0a28_i2c_write(0xfd,0x00);
			sp0a28_i2c_write(0x32,temp|0x05);
			break;
		case CAMERA_ANTIBANDING_50HZ:
			CDBG("--CAMERA--CAMERA_ANTIBANDING_50HZ\n");
			rc = SP0A28Core_WritePREG(sp0a28_antibanding_50z_tbl);
			//sp0a28_i2c_write(0xfd,0x00);
			sp0a28_i2c_write(0x32,temp|0x05);
			break;
		case CAMERA_ANTIBANDING_AUTO:
			CDBG("--CAMERA--CAMERA_ANTIBANDING_AUTO\n");
			rc = SP0A28Core_WritePREG(sp0a28_antibanding_auto_tbl);
			//sp0a28_i2c_write(0xfd,0x00);
			sp0a28_i2c_write(0x32,temp|0x05);
			break;
		default:
			CDBG("--CAMERA--CAMERA_ANTIBANDING_ERROR COMMAND\n");
			break;
	}
	CDBG("--CAMERA-- %s ...(End)\n",__func__);
	return rc;
}

static int sp0a28_set_saturation(int saturation)
{
	long rc = 0;

	CDBG("--CAMERA-- %s ...(Start)\n",__func__);
	CDBG("--CAMERA-- %s ...saturation = %d\n",__func__ , saturation);

	if (effect_value == CAMERA_EFFECT_OFF)
	{
		switch (saturation)
		{
			case CAMERA_SATURATION_LV0:
				CDBG("--CAMERA--CAMERA_SATURATION_LV0\n");
				rc = SP0A28Core_WritePREG(sp0a28_saturation_lv0_tbl);
				break;
			case CAMERA_SATURATION_LV1:
				CDBG("--CAMERA--CAMERA_SATURATION_LV1\n");
				rc = SP0A28Core_WritePREG(sp0a28_saturation_lv1_tbl);
				break;
			case CAMERA_SATURATION_LV2:
				CDBG("--CAMERA--CAMERA_SATURATION_LV2\n");
				rc = SP0A28Core_WritePREG(sp0a28_saturation_lv2_tbl);
				break;
			case CAMERA_SATURATION_LV3:
				CDBG("--CAMERA--CAMERA_SATURATION_LV3\n");
				rc = SP0A28Core_WritePREG(sp0a28_saturation_lv3_tbl);
				break;
			case CAMERA_SATURATION_LV4:
				CDBG("--CAMERA--CAMERA_SATURATION_LV4\n");
				rc = SP0A28Core_WritePREG(sp0a28_saturation_default_lv4_tbl);
				break;
			case CAMERA_SATURATION_LV5:
				CDBG("--CAMERA--CAMERA_SATURATION_LV5\n");
				rc = SP0A28Core_WritePREG(sp0a28_saturation_lv5_tbl);
				break;
			case CAMERA_SATURATION_LV6:
				CDBG("--CAMERA--CAMERA_SATURATION_LV6\n");
				rc = SP0A28Core_WritePREG(sp0a28_saturation_lv6_tbl);
				break;
			case CAMERA_SATURATION_LV7:
				CDBG("--CAMERA--CAMERA_SATURATION_LV7\n");
				rc = SP0A28Core_WritePREG(sp0a28_saturation_lv7_tbl);
				break;
			case CAMERA_SATURATION_LV8:
				CDBG("--CAMERA--CAMERA_SATURATION_LV8\n");
				rc = SP0A28Core_WritePREG(sp0a28_saturation_lv8_tbl);
				break;
			case CAMERA_SATURATION_LV8+1://CAMERA_SATURATION_LV9:
				CDBG("--CAMERA--CAMERA_SATURATION_LV9\n");
				rc = SP0A28Core_WritePREG(sp0a28_saturation_lv9_tbl);
				break;
			case CAMERA_SATURATION_LV8+2://CAMERA_SATURATION_LV10:
				CDBG("--CAMERA--CAMERA_SATURATION_LV10\n");
				rc = SP0A28Core_WritePREG(sp0a28_saturation_lv10_tbl);
				break;
			default:
				CDBG("--CAMERA--CAMERA_SATURATION_ERROR COMMAND\n");
				break;
		}
	}

	//for recover saturation level when change special effect
	switch (saturation)
	{
		case CAMERA_SATURATION_LV0:
			CDBG("--CAMERA--CAMERA_SATURATION_LV0\n");
			SAT_U = 0x00;
			SAT_V = 0x00;
			break;
		case CAMERA_SATURATION_LV1:
			CDBG("--CAMERA--CAMERA_SATURATION_LV1\n");
			SAT_U = 0x0a;
			SAT_V = 0x0a;
			break;
		case CAMERA_SATURATION_LV2:
			CDBG("--CAMERA--CAMERA_SATURATION_LV2\n");
			SAT_U = 0x10;
			SAT_V = 0x10;
			break;
		case CAMERA_SATURATION_LV3:
			CDBG("--CAMERA--CAMERA_SATURATION_LV3\n");
			SAT_U = 0x20;
			SAT_V = 0x20;
			break;
		case CAMERA_SATURATION_LV4:
			CDBG("--CAMERA--CAMERA_SATURATION_LV4\n");
			SAT_U = 0x30;
			SAT_V = 0x30;
			break;
		case CAMERA_SATURATION_LV5:
			CDBG("--CAMERA--CAMERA_SATURATION_LV5\n");
			SAT_U = 0x40;
			SAT_V = 0x40;
			break;
		case CAMERA_SATURATION_LV6:
			CDBG("--CAMERA--CAMERA_SATURATION_LV6\n");
			SAT_U = 0x50;
			SAT_V = 0x50;
			break;
		case CAMERA_SATURATION_LV7:
			CDBG("--CAMERA--CAMERA_SATURATION_LV7\n");
			SAT_U = 0x60;
			SAT_V = 0x60;
			break;
		case CAMERA_SATURATION_LV8:
			CDBG("--CAMERA--CAMERA_SATURATION_LV8\n");
			SAT_U = 0x70;
			SAT_V = 0x70;
			break;
		case CAMERA_SATURATION_LV8+1://CAMERA_SATURATION_LV9:
			CDBG("--CAMERA--CAMERA_SATURATION_LV9\n");
			SAT_U = 0x80;
			SAT_V = 0x80;
			break;
		case CAMERA_SATURATION_LV8+2://CAMERA_SATURATION_LV10:
			CDBG("--CAMERA--CAMERA_SATURATION_LV10\n");
			SAT_U = 0x88;
			SAT_V = 0x88;
			break;
		default:
			CDBG("--CAMERA--CAMERA_SATURATION_ERROR COMMAND\n");
			break;
	}
	CDBG("--CAMERA-- %s ...(End)\n",__func__);
	return rc;
}

static long sp0a28_set_effect(int mode, int effect)
{
	int rc = 0;
	CDBG("--CAMERA-- %s ...(Start)\n",__func__);
	switch (mode)
	{
		case SENSOR_PREVIEW_MODE:
			break;
		case SENSOR_HFR_60FPS_MODE:
			break;
		case SENSOR_HFR_90FPS_MODE:
			//Context A Special Effects /
			CDBG("--CAMERA-- %s ...SENSOR_PREVIEW_MODE\n",__func__);
			break;
		case SENSOR_SNAPSHOT_MODE:
			// Context B Special Effects /
			CDBG("--CAMERA-- %s ...SENSOR_SNAPSHOT_MODE\n",__func__);
			break;
		default:
			break;
	}
	effect_value = effect;
	switch (effect)
	{
		case CAMERA_EFFECT_OFF: {
			CDBG("--CAMERA-- %s ...CAMERA_EFFECT_OFF\n",__func__);
			rc = SP0A28Core_WritePREG(sp0a28_effect_normal_tbl);
			//??sp0a28_i2c_write(0xda, SAT_U); //for recover saturation level when change special effect
			//??sp0a28_i2c_write(0xdb, SAT_V); //for recover saturation level when change special effect
			break;
		}
		case CAMERA_EFFECT_MONO: {
			CDBG("--CAMERA-- %s ...CAMERA_EFFECT_MONO\n",__func__);
			rc = SP0A28Core_WritePREG(sp0a28_effect_mono_tbl);
			break;
		}
		case CAMERA_EFFECT_BW: {
			CDBG("--CAMERA-- %s ...CAMERA_EFFECT_BW\n",__func__);
			rc = SP0A28Core_WritePREG(sp0a28_effect_bw_tbl);
			break;
		}
		case CAMERA_EFFECT_BLUISH: {
			CDBG("--CAMERA-- %s ...CAMERA_EFFECT_BLUISH\n",__func__);
			rc = SP0A28Core_WritePREG(sp0a28_effect_bluish_tbl);
			break;
		}
		case CAMERA_EFFECT_SOLARIZE: {
			CDBG("--CAMERA-- %s ...CAMERA_EFFECT_NEGATIVE(Not Supported)!\n",__func__);
			break;
		}
		case CAMERA_EFFECT_SEPIA: {
			CDBG("--CAMERA-- %s ...CAMERA_EFFECT_SEPIA\n",__func__);
			rc = SP0A28Core_WritePREG(sp0a28_effect_sepia_tbl);
			break;
		}
		case CAMERA_EFFECT_REDDISH: {
			CDBG("--CAMERA-- %s ...CAMERA_EFFECT_REDDISH\n",__func__);
			rc = SP0A28Core_WritePREG(sp0a28_effect_reddish_tbl);
			break;
		}
		case CAMERA_EFFECT_GREENISH: {
			CDBG("--CAMERA-- %s ...CAMERA_EFFECT_GREENISH\n",__func__);
			rc = SP0A28Core_WritePREG(sp0a28_effect_greenish_tbl);
			break;
		}
		case CAMERA_EFFECT_NEGATIVE: {
			CDBG("--CAMERA-- %s ...CAMERA_EFFECT_NEGATIVE\n",__func__);
			rc = SP0A28Core_WritePREG(sp0a28_effect_negative_tbl);
			break;
		}
		default: {
			CDBG("--CAMERA-- %s ...Default(Not Support)\n",__func__);
		}
	}
	sp0a28_effect = effect;
	//Refresh Sequencer /
	CDBG("--CAMERA-- %s ...(End)\n",__func__);
	return rc;
}

static int sp0a28_set_contrast(int contrast)
{
	int rc = 0;
	CDBG("--CAMERA-- %s ...(Start)\n",__func__);
	CDBG("--CAMERA-- %s ...contrast = %d\n",__func__ , contrast);
	if (effect_value == CAMERA_EFFECT_OFF)
	{
		switch (contrast)
		{
		case CAMERA_CONTRAST_LV0:
			CDBG("--CAMERA--CAMERA_CONTRAST_LV0\n");
			rc = SP0A28Core_WritePREG(sp0a28_contrast_lv0_tbl);
			break;
		case CAMERA_CONTRAST_LV1:
			CDBG("--CAMERA--CAMERA_CONTRAST_LV1\n");
			rc = SP0A28Core_WritePREG(sp0a28_contrast_lv1_tbl);
			break;
		case CAMERA_CONTRAST_LV2:
			CDBG("--CAMERA--CAMERA_CONTRAST_LV2\n");
			rc = SP0A28Core_WritePREG(sp0a28_contrast_lv2_tbl);
			break;
		case CAMERA_CONTRAST_LV3:
			CDBG("--CAMERA--CAMERA_CONTRAST_LV3\n");
			rc = SP0A28Core_WritePREG(sp0a28_contrast_lv3_tbl);
			break;
		case CAMERA_CONTRAST_LV4:
			CDBG("--CAMERA--CAMERA_CONTRAST_LV4\n");
			rc = SP0A28Core_WritePREG(sp0a28_contrast_default_lv4_tbl);
			break;
		case CAMERA_CONTRAST_LV5:
			CDBG("--CAMERA--CAMERA_CONTRAST_LV5\n");
			rc = SP0A28Core_WritePREG(sp0a28_contrast_lv5_tbl);
			break;
		case CAMERA_CONTRAST_LV6:
			CDBG("--CAMERA--CAMERA_CONTRAST_LV6\n");
			rc = SP0A28Core_WritePREG(sp0a28_contrast_lv6_tbl);
			break;
		case CAMERA_CONTRAST_LV7:
			CDBG("--CAMERA--CAMERA_CONTRAST_LV7\n");
			rc = SP0A28Core_WritePREG(sp0a28_contrast_lv7_tbl);
			break;
		case CAMERA_CONTRAST_LV8:
			CDBG("--CAMERA--CAMERA_CONTRAST_LV8\n");
			rc = SP0A28Core_WritePREG(sp0a28_contrast_lv8_tbl);
			break;
        case CAMERA_CONTRAST_LV9:
            CDBG("--CAMERA--CAMERA_CONTRAST_LV9\n");
            rc = SP0A28Core_WritePREG(sp0a28_contrast_lv9_tbl);
            break;
        case CAMERA_CONTRAST_LV9+1://CAMERA_CONTRAST_LV10:
            CDBG("--CAMERA--CAMERA_CONTRAST_LV10\n");
            rc = SP0A28Core_WritePREG(sp0a28_contrast_lv10_tbl);
            break;
		default:
			CDBG("--CAMERA--CAMERA_CONTRAST_ERROR COMMAND\n");
			break;
		}
	}
	CDBG("--CAMERA-- %s ...(End)\n",__func__);
	return rc;
}

static int sp0a28_set_sharpness(int sharpness)
{
	int rc = 0;
	CDBG("--CAMERA-- %s ...(Start)\n",__func__);
	CDBG("--CAMERA-- %s ...sharpness = %d\n",__func__ , sharpness);
	if (effect_value == CAMERA_EFFECT_OFF)
	{
		switch(sharpness)
		{
			case CAMERA_SHARPNESS_LV0:
				CDBG("--CAMERA--CAMERA_SHARPNESS_LV0\n");
				rc = SP0A28Core_WritePREG(sp0a28_sharpness_lv0_tbl);
				break;
			case CAMERA_SHARPNESS_LV1:
				CDBG("--CAMERA--CAMERA_SHARPNESS_LV1\n");
				rc = SP0A28Core_WritePREG(sp0a28_sharpness_lv1_tbl);
				break;
			case CAMERA_SHARPNESS_LV2:
				CDBG("--CAMERA--CAMERA_SHARPNESS_LV2\n");
				rc = SP0A28Core_WritePREG(sp0a28_sharpness_default_lv2_tbl);
				break;
			case CAMERA_SHARPNESS_LV3:
				CDBG("--CAMERA--CAMERA_SHARPNESS_LV3\n");
				rc = SP0A28Core_WritePREG(sp0a28_sharpness_lv3_tbl);
				break;
			case CAMERA_SHARPNESS_LV4:
				CDBG("--CAMERA--CAMERA_SHARPNESS_LV4\n");
				rc = SP0A28Core_WritePREG(sp0a28_sharpness_lv4_tbl);
				break;
			case CAMERA_SHARPNESS_LV5:
				CDBG("--CAMERA--CAMERA_SHARPNESS_LV5\n");
				rc = SP0A28Core_WritePREG(sp0a28_sharpness_lv5_tbl);
				break;
			case CAMERA_SHARPNESS_LV6:
				CDBG("--CAMERA--CAMERA_SHARPNESS_LV6\n");
				rc = SP0A28Core_WritePREG(sp0a28_sharpness_lv6_tbl);
				break;
			default:
				CDBG("--CAMERA--CAMERA_SHARPNESS_ERROR COMMAND\n");
				break;
		}
	}
	CDBG("--CAMERA-- %s ...(End)\n",__func__);
	return rc;
}

#if 0
static int sp0a28_set_iso(int8_t iso_type)
{
	long rc = 0;
	uint8_t temp = 0;
	
	sp0a28_i2c_write(0xfd,0x00);
	sp0a28_i2c_read(0x32, &temp, 1);
	CDBG("--CAMERA-- %s ...(Start)\n",__func__);
	CDBG("--CAMERA-- %s ...iso_type = %d\n",__func__ , iso_type);
	CDBG("--CAMERA-- %s ...%x\n",__func__ , temp);
	switch(iso_type)
	{
		case CAMERA_ISO_TYPE_AUTO:
			CDBG("--CAMERA--CAMERA_ISO_TYPE_AUTO\n");
			sp0a28_i2c_write(0x32,temp|0x05);
			rc = SP0A28Core_WritePREG(sp0a28_iso_type_auto);
			CDBG("--CAMERA-- %s ...%x\n",__func__ , temp|0x05);
			break;
		case CAMEAR_ISO_TYPE_HJR:
			CDBG("--CAMERA--CAMEAR_ISO_TYPE_HJR\n");
			sp0a28_i2c_write(0x32,temp|0x05); 
			rc = SP0A28Core_WritePREG(sp0a28_iso_type_auto);
			break;
		case CAMEAR_ISO_TYPE_100:
			CDBG("--CAMERA--CAMEAR_ISO_TYPE_100\n");		
			sp0a28_i2c_write(0x32,temp&(~(0x05)));
			rc = SP0A28Core_WritePREG(sp0a28_iso_type_100);
			break;
		case CAMERA_ISO_TYPE_200:
			CDBG("--CAMERA--CAMERA_ISO_TYPE_200\n");
			sp0a28_i2c_write(0x32,temp&(~(0x05)));
			rc = SP0A28Core_WritePREG(sp0a28_iso_type_200);
			break;
		case CAMERA_ISO_TYPE_400:
			CDBG("--CAMERA--CAMERA_ISO_TYPE_400\n");
			sp0a28_i2c_write(0x32,temp&(~(0x05)));
			rc = SP0A28Core_WritePREG(sp0a28_iso_type_400);
			break;
		case CAMEAR_ISO_TYPE_800:
			CDBG("--CAMERA--CAMEAR_ISO_TYPE_800\n");
			sp0a28_i2c_write(0x32,temp&(~(0x05)));
			rc = SP0A28Core_WritePREG(sp0a28_iso_type_800);
			CDBG("--CAMERA-- XX%s ...%x\n",__func__ , temp&(~(0x05)));
			break;
		case CAMERA_ISO_TYPE_1600:
			CDBG("--CAMERA--CAMERA_ISO_TYPE_1600\n");
			sp0a28_i2c_write(0x32,temp&(~(0x05)));
			rc = SP0A28Core_WritePREG(sp0a28_iso_type_1600);
			CDBG("--CAMERA-- aa %s ...%x\n",__func__ , temp&(~(0x05)));
			break;
		default:
			CDBG("--CAMERA--ERROR ISO TYPE\n");
			break;
	}
	CDBG("--CAMERA-- %s ...(End)\n",__func__);
	return rc;
}
#endif

static int sp0a28_set_wb_oem(uint8_t param)
{
    int rc = 0;
	uint8_t temp = 0;
	
	
	sp0a28_i2c_write(0xfd,0x00);
	sp0a28_i2c_read(0x32, &temp, 1);
    CDBG("--CAMERA--%s runs\r\n",__func__ );

    switch(param)
    {
		case CAMERA_WB_AUTO:
			CDBG("--CAMERA-- CAMERA_WB_AUTO %d\n", param);
			rc = SP0A28Core_WritePREG(sp0a28_wb_auto);
			sp0a28_i2c_write(0xfd,0x00);
			sp0a28_i2c_write(0x32,temp|0x08);
			break;
		case CAMERA_WB_INCANDESCENT:
			CDBG("--CAMERA-- CAMERA_WB_INCANDESCENT %d\n", param);
			//sp0a28_i2c_write(0xfd,0x00);
			sp0a28_i2c_write(0x32,temp&(~(0x08)));
			rc = SP0A28Core_WritePREG(sp0a28_wb_incandescent);
			break;
		case CAMERA_WB_FLUORESCENT:
			CDBG("--CAMERA-- CAMERA_WB_FLUORESCENT %d\n", param);
			sp0a28_i2c_write(0x32,temp&(~(0x08)));
			rc = SP0A28Core_WritePREG(sp0a28_wb_fluorescent);
			break;
		case CAMERA_WB_DAYLIGHT:
			CDBG("--CAMERA-- CAMERA_WB_DAYLIGHT %d\n", param);
			sp0a28_i2c_write(0x32,temp&(~(0x08)));
			rc = SP0A28Core_WritePREG(sp0a28_wb_daylight);
			break;
		case CAMERA_WB_CLOUDY_DAYLIGHT:
			CDBG("--CAMERA-- CAMERA_WB_CLOUDY_DAYLIGHT %d\n", param);
			sp0a28_i2c_write(0x32,temp&(~(0x08)));
			rc = SP0A28Core_WritePREG(sp0a28_wb_cloudy);
			break;
		default:
			CDBG("--CAMERA-- %d\n", param);
			break;
    }
    return rc;
}

static int sp0a28_probe_init_sensor(const struct msm_camera_sensor_info *data)
{
	uint8_t model_id = 0;
	int32_t rc = 0;
	/*The reset pin is not physically connected to the sensor.
	The standby pin will do the reset hence there is no need
	to request the gpio reset*/

	/* Read sensor Model ID: */
	rc = sp0a28_i2c_read(REG_SP0A28_MODEL_ID, &model_id, 1);
	if (rc < 0)
		goto init_probe_fail;

	printk("[sp0a28] model_id = 0x%x\n",model_id);
	/* 4. Compare sensor ID to SP0A28 ID: */
	if (model_id != SP0A28_MODEL_ID) {
		rc = -ENODEV;
		goto init_probe_fail;
	}
	goto init_probe_done;
init_probe_fail:
	pr_warning(" sp0a28_probe_init_sensor fails\n");
init_probe_done:
	CDBG(" sp0a28_probe_init_sensor finishes\n");
	return rc;
}

int sp0a28_sensor_open_init(const struct msm_camera_sensor_info *data)
{
	int32_t rc = 0;

	CDBG("%s: %d\n", __func__, __LINE__);
	CDBG("Calling sp0a28_sensor_open_init\n");
	sp0a28_ctrl = kzalloc(sizeof(struct sp0a28_ctrl_t), GFP_KERNEL);
	if (!sp0a28_ctrl) {
		CDBG("sp0a28_init failed!\n");
		rc = -ENOMEM;
		goto init_done;
	}
	sp0a28_ctrl->fps_divider = 1 * 0x00000400;
	sp0a28_ctrl->pict_fps_divider = 1 * 0x00000400;
	sp0a28_ctrl->fps = 30 * Q8;
	sp0a28_ctrl->set_test = TEST_OFF;
	sp0a28_ctrl->prev_res = QTR_SIZE;
	sp0a28_ctrl->pict_res = FULL_SIZE;
	sp0a28_ctrl->curr_res = INVALID_SIZE;

	if (data)
		sp0a28_ctrl->sensordata = data;

	/* enable mclk first */
	front_camera_power_down();
	msm_camio_clk_rate_set(24000000);
	msleep(20);
	front_camera_power_up();
	msleep(5);
	front_camera_power_reset();
	msleep(5);

	rc = sp0a28_probe_init_sensor(data);
	if (rc < 0) {
		CDBG("Calling sp0a28_sensor_open_init fail\n");
		goto init_fail;
	}

	rc = sp0a28_sensor_setting(REG_INIT, RES_PREVIEW);
	if (rc < 0)
		goto init_fail;
	else
		goto init_done;

init_fail:
	CDBG(" sp0a28_sensor_open_init fail\n");
	kfree(sp0a28_ctrl);
init_done:
	CDBG("sp0a28_sensor_open_init done\n");
	return rc;
}

static int sp0a28_init_client(struct i2c_client *client)
{
	/* Initialize the MSM_CAMI2C Chip */
	init_waitqueue_head(&sp0a28_wait_queue);
	return 0;
}

static const struct i2c_device_id sp0a28_i2c_id[] = {
    {"sp0a28", 0},
    { }
};
#if 0
static ssize_t sp0a28_set_settle_cnt(struct device* dev,                                                            
							struct device_attribute* attr, const char* buf, size_t count) 
{                                                                                                                    
	uint32_t set_cnt=0;

	printk(KERN_INFO "enter %s buf=%s\n", __func__, buf);																													 
	if(strlen(buf)==0)
	{
		printk(KERN_INFO "string length is Zero!\n");
		return 0;
	}
	set_cnt= simple_strtoul(buf, NULL, 16);
	printk("settle_cnt=0x%x\n", set_cnt);

	settle_cnt_val=(uint8_t)set_cnt;
	printk("the settle_cnt_val=0x%x\n",settle_cnt_val );
                                                                                                             
	return	count;                                                                                                
} 
static DEVICE_ATTR(set_settle_cnt, 0666, NULL, sp0a28_set_settle_cnt); 
#endif
static int sp0a28_i2c_probe(struct i2c_client *client,
    const struct i2c_device_id *id)
{
    int rc = 0;
    CDBG("sp0a28_i2c_probe called!\n");

    if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
        CDBG("i2c_check_functionality failed\n");
        goto probe_failure;
    }

    sp0a28_sensorw = kzalloc(sizeof(struct sp0a28_work_t), GFP_KERNEL);
    if (!sp0a28_sensorw) {
        CDBG("kzalloc failed.\n");
        rc = -ENOMEM;
        goto probe_failure;
    }

    i2c_set_clientdata(client, sp0a28_sensorw);
    sp0a28_init_client(client);
    sp0a28_client = client;


    CDBG("sp0a28_i2c_probe success! rc = %d\n", rc);
    return 0;

probe_failure:
    CDBG("sp0a28_i2c_probe failed! rc = %d\n", rc);
    return rc;
}

static int __exit sp0a28_remove(struct i2c_client *client)
{
    struct sp0a28_work_t_t *sensorw = i2c_get_clientdata(client);
    free_irq(client->irq, sensorw);
    sp0a28_client = NULL;
    kfree(sensorw);
    return 0;
}

static struct i2c_driver sp0a28_i2c_driver = {
    .id_table = sp0a28_i2c_id,
    .probe  = sp0a28_i2c_probe,
    .remove = __exit_p(sp0a28_i2c_remove),
    .driver = {
        .name = "sp0a28",
    },
};

int sp0a28_sensor_config(void __user *argp)
{
    struct sensor_cfg_data cdata;
    long rc = 0;
    if (copy_from_user(&cdata, (void *)argp, sizeof(struct sensor_cfg_data)))
		return -EFAULT;
    mutex_lock(&sp0a28_mut);
    CDBG("sp0a28_sensor_config: cfgtype = %d\n", cdata.cfgtype);
    switch (cdata.cfgtype)
    {
		case CFG_SET_MODE:
			rc = sp0a28_set_sensor_mode(cdata.mode, cdata.rs);
			break;
		case CFG_SET_EFFECT:
			CDBG("--CAMERA-- CFG_SET_EFFECT mode=%d, effect = %d !!\n",cdata.mode, cdata.cfg.effect);
			rc = sp0a28_set_effect(cdata.mode, cdata.cfg.effect);
			break;
		case CFG_START:
			CDBG("--CAMERA-- CFG_START (Not Support) !!\n");
			// Not Support
			break;
		case CFG_PWR_UP:
			CDBG("--CAMERA-- CFG_PWR_UP (Not Support) !!\n");
			// Not Support
			break;
		case CFG_PWR_DOWN:
			CDBG("--CAMERA-- CFG_PWR_DOWN !!\n");
			front_camera_power_down();
			break;
		case CFG_WRITE_EXPOSURE_GAIN:
			CDBG("--CAMERA-- CFG_WRITE_EXPOSURE_GAIN (Not Support) !!\n");
			// Not Support
			break;
		case CFG_SET_DEFAULT_FOCUS:
		    CDBG("--CAMERA-- CFG_SET_DEFAULT_FOCUS (Not Implement) !!\n");
		    break;
		case CFG_MOVE_FOCUS:
		    CDBG("--CAMERA-- CFG_MOVE_FOCUS (Not Implement) !!\n");
		    break;
		case CFG_REGISTER_TO_REAL_GAIN:
			CDBG("--CAMERA-- CFG_REGISTER_TO_REAL_GAIN (Not Support) !!\n");
			// Not Support
			break;
		case CFG_REAL_TO_REGISTER_GAIN:
			CDBG("--CAMERA-- CFG_REAL_TO_REGISTER_GAIN (Not Support) !!\n");
			// Not Support
			break;
		case CFG_SET_FPS:
			CDBG("--CAMERA-- CFG_SET_FPS (Not Support) !!\n");
			// Not Support
			break;
		case CFG_SET_PICT_FPS:
			CDBG("--CAMERA-- CFG_SET_PICT_FPS (Not Support) !!\n");
			// Not Support
			break;
		case CFG_SET_BRIGHTNESS:
		    CDBG("--CAMERA-- CFG_SET_BRIGHTNESS  !!\n");
		    //rc = sp0a28_set_brightness(cdata.cfg.brightness);
		    break;
		case CFG_SET_CONTRAST:
		    CDBG("--CAMERA-- CFG_SET_CONTRAST  !!\n");
		    rc = sp0a28_set_contrast(cdata.cfg.contrast);
		    break;
		case CFG_SET_ZOOM:
			CDBG("--CAMERA-- CFG_SET_ZOOM (Not Support) !!\n");
			// Not Support
			break;
		case CFG_SET_EXPOSURE_MODE:
		    CDBG("--CAMERA-- CFG_SET_EXPOSURE_MODE !!\n");
		    //rc = sp0a28_set_exposure_mode(cdata.cfg.ae_mode);
		    break;
		case CFG_SET_WB:
			CDBG("--CAMERA-- CFG_SET_WB!!\n");
			sp0a28_set_wb_oem(cdata.cfg.wb_val);
			rc = 0 ;
			break;
		case CFG_SET_ANTIBANDING:
		    CDBG("--CAMERA-- CFG_SET_ANTIBANDING antibanding = %d!!\n", cdata.cfg.antibanding);
		    rc = sp0a28_set_antibanding(cdata.cfg.antibanding);
		    break;
		case CFG_SET_EXP_GAIN:
			CDBG("--CAMERA-- CFG_SET_EXP_GAIN (Not Support) !!\n");
			// Not Support
			break;
		case CFG_SET_PICT_EXP_GAIN:
			CDBG("--CAMERA-- CFG_SET_PICT_EXP_GAIN (Not Support) !!\n");
			// Not Support
			break;
		case CFG_SET_LENS_SHADING:
		    CDBG("--CAMERA-- CFG_SET_LENS_SHADING !!\n");
		    //rc = sp0a28_lens_shading_enable(cdata.cfg.lens_shading);
		    break;
		case CFG_GET_PICT_FPS:
			CDBG("--CAMERA-- CFG_GET_PICT_FPS (Not Support) !!\n");
			// Not Support
			break;
		case CFG_GET_PREV_L_PF:
			CDBG("--CAMERA-- CFG_GET_PREV_L_PF (Not Support) !!\n");
			// Not Support
			break;
		case CFG_GET_PREV_P_PL:
			CDBG("--CAMERA-- CFG_GET_PREV_P_PL (Not Support) !!\n");
			// Not Support
			break;
		case CFG_GET_PICT_L_PF:
			CDBG("--CAMERA-- CFG_GET_PICT_L_PF (Not Support) !!\n");
			// Not Support
			break;
		case CFG_GET_PICT_P_PL:
			CDBG("--CAMERA-- CFG_GET_PICT_P_PL (Not Support) !!\n");
			// Not Support
			break;
		case CFG_GET_AF_MAX_STEPS:
			CDBG("--CAMERA-- CFG_GET_AF_MAX_STEPS (Not Support) !!\n");
			// Not Support
			break;
		case CFG_GET_PICT_MAX_EXP_LC:
			CDBG("--CAMERA-- CFG_GET_PICT_MAX_EXP_LC (Not Support) !!\n");
			// Not Support
			break;
		case CFG_SEND_WB_INFO:
			CDBG("--CAMERA-- CFG_SEND_WB_INFO (Not Support) !!\n");
			// Not Support
			break;
		case CFG_SENSOR_INIT:
			CDBG("--CAMERA-- CFG_SENSOR_INIT (Not Support) !!\n");
			// Not Support
			break;
		case CFG_SET_SATURATION:
		    CDBG("--CAMERA-- CFG_SET_SATURATION !!\n");
		    rc = sp0a28_set_saturation(cdata.cfg.saturation);
		    break;
		case CFG_SET_SHARPNESS:
		    CDBG("--CAMERA-- CFG_SET_SHARPNESS !!\n");
		    rc = sp0a28_set_sharpness(cdata.cfg.sharpness);
		    break;
		case CFG_SET_TOUCHAEC:
		    CDBG("--CAMERA-- CFG_SET_TOUCHAEC!!\n");
		    //sp0a28_set_touchaec(cdata.cfg.aec_cord.x, cdata.cfg.aec_cord.y);
		    rc = 0 ;
		    break;
		case CFG_SET_AUTO_FOCUS:
		    CDBG("--CAMERA-- CFG_SET_AUTO_FOCUS (Not Support) !!\n");
		    // Not Support
		    break;
		case CFG_SET_AUTOFLASH:
		    CDBG("--CAMERA-- CFG_SET_AUTOFLASH (Not Support) !!\n");
		    // Not Support
		    break;
		case CFG_SET_EXPOSURE_COMPENSATION:
		    CDBG("--CAMERA-- CFG_SET_EXPOSURE_COMPENSATION !\n");
		    rc = sp0a28_set_exposure_compensation(cdata.cfg.exp_compensation);
		    break;
#if 0
		case CFG_SET_ISO:
			CDBG("--CAMERA-- CFG_SET_ISO !\n");
			rc = sp0a28_set_iso(cdata.cfg.iso_type);
		    break;
#endif
		default:
		    CDBG("--CAMERA-- %s: Command=%d (Not Implement)!!\n", __func__, cdata.cfgtype);
		    rc = -EINVAL;
		    break;
	}
	mutex_unlock(&sp0a28_mut);
	return rc;
}
static int sp0a28_sensor_release(void)
{
    int rc = -EBADF;
    mutex_lock(&sp0a28_mut);
    front_camera_power_down();
    kfree(sp0a28_ctrl);
    sp0a28_ctrl = NULL;
    CDBG("sp0a28_release completed\n");
    mutex_unlock(&sp0a28_mut);

    return rc;
}

static int sp0a28_sensor_probe(const struct msm_camera_sensor_info *info,
        struct msm_sensor_ctrl *s)
{
    int rc = 0;

	printk("[sp0a28] enter %s\n", __func__);
    rc = i2c_add_driver(&sp0a28_i2c_driver);
    if (rc < 0 || sp0a28_client == NULL) {
        rc = -ENOTSUPP;
        goto probe_fail;
    }
	#if 0
	int res=0;
	res = device_create_file(&sp0a28_client->dev, &dev_attr_set_settle_cnt);											 
	if(res){																									  
		printk("[sp0a28:] device create file dev_attr_set_settle_cnt failed!\n");																																			
	} 
	#endif

    front_camera_power_down();
    msm_camio_clk_rate_set(24000000);
    msleep(20);
    front_camera_power_up();
    msleep(5);
	front_camera_power_reset();
	msleep(5);

    rc = sp0a28_probe_init_sensor(info);
    if (rc < 0)
        goto probe_fail;

    s->s_init = sp0a28_sensor_open_init;
    s->s_release = sp0a28_sensor_release;
    s->s_config  = sp0a28_sensor_config;
    s->s_camera_type = FRONT_CAMERA_2D;
    s->s_mount_angle = info->sensor_platform_info->mount_angle;
	
	printk("[sp0a28] end %s\n", __func__);

    return rc;

probe_fail:
    CDBG("sp0a28_sensor_probe: SENSOR PROBE FAILS!\n");
    i2c_del_driver(&sp0a28_i2c_driver);
    return rc;
}

static int __sp0a28_probe(struct platform_device *pdev)
{
    return msm_camera_drv_start(pdev, sp0a28_sensor_probe);
}

static struct platform_driver msm_camera_driver = {
    .probe = __sp0a28_probe,
    .driver = {
        .name = "msm_camera_sp0a28",
        .owner = THIS_MODULE,
    },
};

static int __init sp0a28_init(void)
{
    return platform_driver_register(&msm_camera_driver);
}

module_init(sp0a28_init);
//late_initcall(sp0a28_init);

MODULE_DESCRIPTION("SP VGA YUV sensor driver");
MODULE_LICENSE("GPL v2");
